@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2015 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -60,6 +60,10 @@ NOEXPORT int parse_socket_error(CLI *, const char *);
 NOEXPORT void print_cipher(CLI *);
 NOEXPORT void auth_user(CLI *, char *);
 NOEXPORT SOCKET connect_local(CLI *);
+#if !defined(USE_WIN32) && !defined(__vms)
+NOEXPORT char **env_alloc(CLI *);
+NOEXPORT void env_free(char **);
+#endif
 NOEXPORT SOCKET connect_remote(CLI *);
 NOEXPORT void connect_cache(SSL_SESSION *, SOCKADDR_UNION *);
 NOEXPORT unsigned connect_index(CLI *);
@@ -70,6 +74,7 @@ NOEXPORT void reset(SOCKET, char *);
 
 /* allocate local data structure for the new thread */
 CLI *alloc_client_session(SERVICE_OPTIONS *opt, SOCKET rfd, SOCKET wfd) {
+    static unsigned long long seq=0;
     CLI *c;
 
     c=str_alloc_detached(sizeof(CLI));
@@ -77,6 +82,7 @@ CLI *alloc_client_session(SERVICE_OPTIONS *opt, SOCKET rfd, SOCKET wfd) {
     c->local_rfd.fd=rfd;
     c->local_wfd.fd=wfd;
     c->redirect=REDIRECT_OFF;
+    c->seq=seq++;
     return c;
 }
 
@@ -145,9 +151,9 @@ NOEXPORT void client_run(CLI *c) {
 #endif
 
 #ifndef USE_FORK
-    enter_critical_section(CRIT_CLIENTS);
+    CRYPTO_w_lock(stunnel_locks[LOCK_CLIENTS]);
     ui_clients(++num_clients);
-    leave_critical_section(CRIT_CLIENTS);
+    CRYPTO_w_unlock(stunnel_locks[LOCK_CLIENTS]);
 #endif
 
         /* initialize the client context */
@@ -165,7 +171,7 @@ NOEXPORT void client_run(CLI *c) {
         c->ssl_wfd=&(c->local_wfd);
     }
     c->fds=s_poll_alloc();
-    addrlist_clear(&c->connect_addr);
+    addrlist_clear(&c->connect_addr, 0);
 
         /* try to process the request */
     err=setjmp(c->err);
@@ -199,7 +205,8 @@ NOEXPORT void client_run(CLI *c) {
         if(rst && c->remote_fd.is_socket) /* reset */
             reset(c->remote_fd.fd, "linger (remote)");
         closesocket(c->remote_fd.fd);
-        s_log(LOG_DEBUG, "Remote socket (FD=%d) closed", c->remote_fd.fd);
+        s_log(LOG_DEBUG, "Remote descriptor (FD=%ld) closed",
+            (long)c->remote_fd.fd);
         c->remote_fd.fd=INVALID_SOCKET;
     }
 
@@ -209,7 +216,8 @@ NOEXPORT void client_run(CLI *c) {
             if(rst && c->local_rfd.is_socket)
                 reset(c->local_rfd.fd, "linger (local)");
             closesocket(c->local_rfd.fd);
-            s_log(LOG_DEBUG, "Local socket (FD=%d) closed", c->local_rfd.fd);
+            s_log(LOG_DEBUG, "Local descriptor (FD=%ld) closed",
+                (long)c->local_rfd.fd);
         } else { /* stdin/stdout */
             if(rst && c->local_rfd.is_socket)
                 reset(c->local_rfd.fd, "linger (local_rfd)");
@@ -226,10 +234,10 @@ NOEXPORT void client_run(CLI *c) {
         child_status(); /* null SIGCHLD handler was used */
     s_log(LOG_DEBUG, "Service [%s] finished", c->opt->servname);
 #else
-    enter_critical_section(CRIT_CLIENTS);
+    CRYPTO_w_lock(stunnel_locks[LOCK_CLIENTS]);
     ui_clients(--num_clients);
     num_clients_copy=num_clients; /* to move s_log() away from CRIT_CLIENTS */
-    leave_critical_section(CRIT_CLIENTS);
+    CRYPTO_w_unlock(stunnel_locks[LOCK_CLIENTS]);
     s_log(LOG_DEBUG, "Service [%s] finished (%ld left)",
         c->opt->servname, num_clients_copy);
 #endif
@@ -340,10 +348,18 @@ NOEXPORT void remote_start(CLI *c) {
     else
         c->remote_fd.fd=connect_remote(c);
 
-    c->remote_fd.is_socket=1; /* always! */
-    s_log(LOG_DEBUG, "Remote socket (FD=%d) initialized", c->remote_fd.fd);
-    if(set_socket_options(c->remote_fd.fd, 2))
-        s_log(LOG_WARNING, "Failed to set remote socket options");
+#ifndef USE_WIN32
+    if(c->opt->option.pty) { /* descriptor created with pty_allocate() */
+        c->remote_fd.is_socket=0;
+    } else
+#endif
+    {
+        c->remote_fd.is_socket=1;
+        if(set_socket_options(c->remote_fd.fd, 2))
+            s_log(LOG_WARNING, "Failed to set remote socket options");
+    }
+    s_log(LOG_DEBUG, "Remote descriptor (FD=%ld) initialized",
+        (long)c->remote_fd.fd);
 }
 
 NOEXPORT void ssl_start(CLI *c) {
@@ -369,9 +385,9 @@ NOEXPORT void ssl_start(CLI *c) {
         }
 #endif
         if(c->opt->session) {
-            enter_critical_section(CRIT_SESSION);
+            CRYPTO_w_lock(stunnel_locks[LOCK_SESSION]);
             SSL_set_session(c->ssl, c->opt->session);
-            leave_critical_section(CRIT_SESSION);
+            CRYPTO_w_unlock(stunnel_locks[LOCK_SESSION]);
         }
         SSL_set_fd(c->ssl, (int)c->remote_fd.fd);
         SSL_set_connect_state(c->ssl);
@@ -386,8 +402,9 @@ NOEXPORT void ssl_start(CLI *c) {
         SSL_set_accept_state(c->ssl);
     }
 
-    unsafe_openssl=SSLeay()<0x0090810fL ||
-        (SSLeay()>=0x10000000L && SSLeay()<0x1000002fL);
+    unsafe_openssl=OpenSSL_version_num()<0x0090810fL ||
+        (OpenSSL_version_num()>=0x10000000L &&
+        OpenSSL_version_num()<0x1000002fL);
     while(1) {
         /* critical section for OpenSSL version < 0.9.8p or 1.x.x < 1.0.0b *
          * this critical section is a crude workaround for CVE-2010-3864   *
@@ -395,7 +412,7 @@ NOEXPORT void ssl_start(CLI *c) {
          * alternative solution is to disable internal session caching     *
          * NOTE: this critical section also covers callbacks (e.g. OCSP)   */
         if(unsafe_openssl)
-            enter_critical_section(CRIT_SSL);
+            CRYPTO_w_lock(stunnel_locks[LOCK_SSL]);
 
         if(c->opt->option.client)
             i=SSL_connect(c->ssl);
@@ -403,7 +420,7 @@ NOEXPORT void ssl_start(CLI *c) {
             i=SSL_accept(c->ssl);
 
         if(unsafe_openssl)
-            leave_critical_section(CRIT_SSL);
+            CRYPTO_w_unlock(stunnel_locks[LOCK_SSL]);
 
         err=SSL_get_error(c->ssl, i);
         if(err==SSL_ERROR_NONE)
@@ -466,10 +483,10 @@ NOEXPORT void ssl_start(CLI *c) {
         SSL_SESSION_set_ex_data(SSL_get_session(c->ssl),
             index_redirect, (void *)c->redirect);
         if(c->opt->option.client) {
-            enter_critical_section(CRIT_SESSION);
+            CRYPTO_w_lock(stunnel_locks[LOCK_SESSION]);
             old_session=c->opt->session;
             c->opt->session=SSL_get1_session(c->ssl); /* store it */
-            leave_critical_section(CRIT_SESSION);
+            CRYPTO_w_unlock(stunnel_locks[LOCK_SESSION]);
             if(old_session)
                 SSL_SESSION_free(old_session); /* release the old one */
         } else { /* SSL server */
@@ -601,10 +618,36 @@ NOEXPORT void transfer(CLI *c) {
         ssl_can_rd=s_poll_canread(c->fds, c->ssl_rfd->fd);
         ssl_can_wr=s_poll_canwrite(c->fds, c->ssl_wfd->fd);
 
-        /****************************** checks for internal failures */
-        if(!(sock_can_rd || sock_can_wr || ssl_can_rd || ssl_can_wr))
-            s_log(LOG_NOTICE,
-                "s_poll_wait returned %d, but no descriptor is ready", err);
+        /****************************** hangups without read or write */
+        if(!(sock_can_rd || sock_can_wr || ssl_can_rd || ssl_can_wr)) {
+            if(s_poll_hup(c->fds, c->sock_rfd->fd) ||
+                    s_poll_hup(c->fds, c->sock_wfd->fd)) {
+                if(c->ssl_ptr) {
+                    s_log(LOG_ERR,
+                        "Socket closed (HUP) with %ld unsent byte(s)",
+                        (long)c->ssl_ptr);
+                    longjmp(c->err, 1); /* reset the sockets */
+                }
+                s_log(LOG_INFO, "Socket closed (HUP)");
+                sock_open_rd=sock_open_wr=0;
+            } else if(s_poll_hup(c->fds, c->ssl_rfd->fd) ||
+                    s_poll_hup(c->fds, c->ssl_wfd->fd)) {
+                if(c->sock_ptr) {
+                    s_log(LOG_ERR,
+                        "SSL socket closed (HUP) with %ld unsent byte(s)",
+                        (long)c->sock_ptr);
+                    longjmp(c->err, 1); /* reset the sockets */
+                }
+                s_log(LOG_INFO, "SSL socket closed (HUP)");
+                SSL_set_shutdown(c->ssl,
+                    SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+            } else { /* please report it to the stunnel-users mailing list */
+                s_log(LOG_ERR, "INTERNAL ERROR: "
+                    "s_poll_wait returned %d, but no descriptor is ready", err);
+                s_poll_dump(c->fds, LOG_ERR);
+                longjmp(c->err, 1);
+            }
+        }
 
         if(c->reneg_state==RENEG_DETECTED && !c->opt->option.renegotiation) {
             s_log(LOG_ERR, "Aborting due to renegotiation request");
@@ -730,8 +773,8 @@ NOEXPORT void transfer(CLI *c) {
                 if(c->sock_ptr) { /* TODO: what about buffered data? */
                     s_log(LOG_ERR,
                         "SSL socket closed (SSL_write) with %ld unsent byte(s)",
-                        c->sock_ptr);
-                    longjmp(c->err, 1); /* reset the socket */
+                        (long)c->sock_ptr);
+                    longjmp(c->err, 1); /* reset the sockets */
                 }
                 s_log(LOG_INFO, "SSL socket closed (SSL_write)");
                 SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
@@ -787,8 +830,8 @@ NOEXPORT void transfer(CLI *c) {
                 if(c->sock_ptr || write_wants_write) {
                     s_log(LOG_ERR,
                         "SSL socket closed (SSL_read) with %ld unsent byte(s)",
-                        c->sock_ptr);
-                    longjmp(c->err, 1); /* reset the socket */
+                        (long)c->sock_ptr);
+                    longjmp(c->err, 1); /* reset the sockets */
                 }
                 s_log(LOG_INFO, "SSL socket closed (SSL_read)");
                 SSL_set_shutdown(c->ssl,
@@ -821,8 +864,8 @@ NOEXPORT void transfer(CLI *c) {
             if(c->ssl_ptr) {
                 s_log(LOG_ERR,
                     "Write socket closed (write hangup) with %ld unsent byte(s)",
-                    c->ssl_ptr);
-                longjmp(c->err, 1); /* reset the socket */
+                    (long)c->ssl_ptr);
+                longjmp(c->err, 1); /* reset the sockets */
             }
             s_log(LOG_INFO, "Write socket closed (write hangup)");
             sock_open_wr=0;
@@ -842,8 +885,8 @@ NOEXPORT void transfer(CLI *c) {
             if(c->sock_ptr || write_wants_write) {
                 s_log(LOG_ERR,
                     "SSL socket closed (write hangup) with %ld unsent byte(s)",
-                    c->sock_ptr);
-                longjmp(c->err, 1); /* reset the socket */
+                    (long)c->sock_ptr);
+                longjmp(c->err, 1); /* reset the sockets */
             }
             s_log(LOG_INFO, "SSL socket closed (write hangup)");
             SSL_set_shutdown(c->ssl,
@@ -906,7 +949,8 @@ NOEXPORT void transfer(CLI *c) {
                 shutdown_wants_read ? "Y" : "n",
                 shutdown_wants_write ? "Y" : "n");
             s_log(LOG_ERR, "socket input buffer: %ld byte(s), "
-                "ssl input buffer: %ld byte(s)", c->sock_ptr, c->ssl_ptr);
+                "ssl input buffer: %ld byte(s)",
+                (long)c->sock_ptr, (long)c->ssl_ptr);
             longjmp(c->err, 1);
         }
 
@@ -1092,10 +1136,11 @@ NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
 
 #else /* standard Unix version */
 
+extern char **environ;
+
 NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
-    char *name, host[40], port[6];
     int fd[2], pid;
-    X509 *peer_cert;
+    char **env;
 #ifdef HAVE_PTHREAD_SIGMASK
     sigset_t newmask;
 #endif
@@ -1109,56 +1154,28 @@ NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
     } else
         if(make_sockets(fd))
             longjmp(c->err, 1);
+    set_nonblock(fd[1], 0); /* switch back to the blocking mode */
 
+    env=env_alloc(c);
     pid=fork();
     c->pid=(unsigned long)pid;
     switch(pid) {
     case -1:    /* error */
         closesocket(fd[0]);
         closesocket(fd[1]);
+        env_free(env);
         ioerror("fork");
         longjmp(c->err, 1);
     case  0:    /* child */
-        tls_alloc(NULL, c->tls, NULL); /* reuse thread-local storage */
+        /* the child is not allowed to play with thread-local storage */
+        /* see http://linux.die.net/man/3/pthread_atfork for details */
         closesocket(fd[0]);
-        set_nonblock(fd[1], 0); /* switch back to blocking mode */
         /* dup2() does not copy FD_CLOEXEC flag */
         dup2(fd[1], 0);
         dup2(fd[1], 1);
-        if(!global_options.option.foreground)
+        if(!global_options.option.log_stderr)
             dup2(fd[1], 2);
         closesocket(fd[1]); /* not really needed due to FD_CLOEXEC */
-
-        if(!getnameinfo(&c->peer_addr.sa, c->peer_addr_len,
-                host, 40, port, 6, NI_NUMERICHOST|NI_NUMERICSERV)) {
-            /* just don't set these variables if getnameinfo() fails */
-            putenv(str_printf("REMOTE_HOST=%s", host));
-            putenv(str_printf("REMOTE_PORT=%s", port));
-            if(c->opt->option.transparent_src) {
-#ifndef LIBDIR
-#define LIBDIR "."
-#endif
-#ifdef MACH64
-                putenv("LD_PRELOAD_32=" LIBDIR "/libstunnel.so");
-                putenv("LD_PRELOAD_64=" LIBDIR "/" MACH64 "/libstunnel.so");
-#elif __osf /* for Tru64 _RLD_LIST is used instead */
-                putenv("_RLD_LIST=" LIBDIR "/libstunnel.so:DEFAULT");
-#else
-                putenv("LD_PRELOAD=" LIBDIR "/libstunnel.so");
-#endif
-            }
-        }
-
-        if(c->ssl) {
-            peer_cert=SSL_get_peer_certificate(c->ssl);
-            if(peer_cert) {
-                name=X509_NAME2text(X509_get_subject_name(peer_cert));
-                putenv(str_printf("SSL_CLIENT_DN=%s", name));
-                name=X509_NAME2text(X509_get_issuer_name(peer_cert));
-                putenv(str_printf("SSL_CLIENT_I_DN=%s", name));
-                X509_free(peer_cert);
-            }
-        }
 #ifdef HAVE_PTHREAD_SIGMASK
         sigemptyset(&newmask);
         sigprocmask(SIG_SETMASK, &newmask, NULL);
@@ -1170,16 +1187,79 @@ NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
         signal(SIGTERM, SIG_DFL);
         signal(SIGQUIT, SIG_DFL);
         signal(SIGINT, SIG_DFL);
-        execvp(c->opt->exec_name, c->opt->exec_args);
-        ioerror(c->opt->exec_name); /* execvp failed */
-        _exit(1);
+        execve(c->opt->exec_name, c->opt->exec_args, env);
+        _exit(1); /* failed, but there is no way to report an error here */
     default: /* parent */
-        s_log(LOG_INFO, "Local mode child started (PID=%lu)", c->pid);
         closesocket(fd[1]);
+        env_free(env);
+        s_log(LOG_INFO, "Local mode child started (PID=%lu)", c->pid);
         return fd[0];
     }
 }
 
+char **env_alloc(CLI *c) {
+    char **env=NULL, **p;
+    unsigned n=0; /* (n+2) keeps the list NULL-terminated */
+    char *name, host[40], port[6];
+    X509 *peer_cert;
+
+    if(!getnameinfo(&c->peer_addr.sa, c->peer_addr_len,
+            host, 40, port, 6, NI_NUMERICHOST|NI_NUMERICSERV)) {
+        /* just don't set these variables if getnameinfo() fails */
+        env=str_realloc(env, (n+2)*sizeof(char *));
+        env[n++]=str_printf("REMOTE_HOST=%s", host);
+        env=str_realloc(env, (n+2)*sizeof(char *));
+        env[n++]=str_printf("REMOTE_PORT=%s", port);
+        if(c->opt->option.transparent_src) {
+#ifndef LIBDIR
+#define LIBDIR "."
+#endif
+#ifdef MACH64
+            env=str_realloc(env, (n+2)*sizeof(char *));
+            env[n++]=str_dup("LD_PRELOAD_32=" LIBDIR "/libstunnel.so");
+            env=str_realloc(env, (n+2)*sizeof(char *));
+            env[n++]=str_dup("LD_PRELOAD_64=" LIBDIR "/" MACH64 "/libstunnel.so");
+#elif __osf /* for Tru64 _RLD_LIST is used instead */
+            env=str_realloc(env, (n+2)*sizeof(char *));
+            env[n++]=str_dup("_RLD_LIST=" LIBDIR "/libstunnel.so:DEFAULT");
+#else
+            env=str_realloc(env, (n+2)*sizeof(char *));
+            env[n++]=str_dup("LD_PRELOAD=" LIBDIR "/libstunnel.so");
+#endif
+        }
+    }
+
+    if(c->ssl) {
+        peer_cert=SSL_get_peer_certificate(c->ssl);
+        if(peer_cert) {
+            name=X509_NAME2text(X509_get_subject_name(peer_cert));
+            env=str_realloc(env, (n+2)*sizeof(char *));
+            env[n++]=str_printf("SSL_CLIENT_DN=%s", name);
+            str_free(name);
+            name=X509_NAME2text(X509_get_issuer_name(peer_cert));
+            env=str_realloc(env, (n+2)*sizeof(char *));
+            env[n++]=str_printf("SSL_CLIENT_I_DN=%s", name);
+            str_free(name);
+            X509_free(peer_cert);
+        }
+    }
+
+    for(p=environ; *p; ++p) {
+        env=str_realloc(env, (n+2)*sizeof(char *));
+        env[n++]=str_dup(*p);
+    }
+
+    return env;
+}
+
+void env_free(char **env) {
+    char **p;
+
+    for(p=env; *p; ++p)
+        str_free(*p);
+    str_free(env);
+}
+
 #endif /* not USE_WIN32 or __vms */
 
 /* connect remote host */
@@ -1188,11 +1268,16 @@ NOEXPORT SOCKET connect_remote(CLI *c) {
     unsigned ind_start, ind_try, ind_cur;
 
     setup_connect_addr(c);
-    if(!c->connect_addr.num) {
+    switch(c->connect_addr.num) {
+    case 0:
         s_log(LOG_ERR, "No remote host resolved");
         longjmp(c->err, 1);
+    case 1:
+        ind_start=0;
+        break;
+    default:
+        ind_start=connect_index(c);
     }
-    ind_start=connect_index(c);
 
     /* try to connect each host from the list */
     for(ind_try=0; ind_try<c->connect_addr.num; ind_try++) {
@@ -1236,10 +1321,10 @@ NOEXPORT void connect_cache(SSL_SESSION *sess, SOCKADDR_UNION *cur_addr) {
     s_log(LOG_INFO, "persistence: %s cached", addr_txt);
     str_free(addr_txt);
 
-    enter_critical_section(CRIT_ADDR);
+    CRYPTO_w_lock(stunnel_locks[LOCK_ADDR]);
     old_addr=SSL_SESSION_get_ex_data(sess, index_addr);
     SSL_SESSION_set_ex_data(sess, index_addr, new_addr);
-    leave_critical_section(CRIT_ADDR);
+    CRYPTO_w_unlock(stunnel_locks[LOCK_ADDR]);
     str_free(old_addr); /* NULL pointers are ignored */
 }
 
@@ -1250,12 +1335,12 @@ NOEXPORT unsigned connect_index(CLI *c) {
     char *addr_txt;
 
     if(c->ssl && SSL_session_reused(c->ssl)) {
-        enter_critical_section(CRIT_ADDR);
+        CRYPTO_r_lock(stunnel_locks[LOCK_ADDR]);
         ptr=SSL_SESSION_get_ex_data(SSL_get_session(c->ssl), index_addr);
         if(ptr) {
             len=addr_len(ptr);
             memcpy(&addr, ptr, (size_t)len);
-            leave_critical_section(CRIT_ADDR);
+            CRYPTO_r_unlock(stunnel_locks[LOCK_ADDR]);
             /* address was copied, ptr itself is no longer valid */
             for(i=0; i<c->connect_addr.num; ++i) {
                 if(addr_len(&c->connect_addr.addr[i])==len &&
@@ -1271,26 +1356,24 @@ NOEXPORT unsigned connect_index(CLI *c) {
             s_log(LOG_INFO, "persistence: %s not available", addr_txt);
             str_free(addr_txt);
         } else {
-            leave_critical_section(CRIT_ADDR);
+            CRYPTO_r_unlock(stunnel_locks[LOCK_ADDR]);
             s_log(LOG_NOTICE, "persistence: No cached address found");
         }
     }
-    i=*c->connect_addr.rr_ptr;
-    /* the race condition here can be safely ignored */
+
     if(c->opt->failover==FAILOVER_RR) {
+        /* the race condition here can be safely ignored */
+        i=*c->connect_addr.rr_ptr;
         *c->connect_addr.rr_ptr=(i+1)%c->connect_addr.num;
-        s_log(LOG_INFO, "failover: round-robin");
+        s_log(LOG_INFO, "failover: round-robin, starting at entry #%d", i);
     } else {
-        s_log(LOG_INFO, "failover: priority");
+        i=0;
+        s_log(LOG_INFO, "failover: priority, starting at entry #0");
     }
     return i;
 }
 
 NOEXPORT void setup_connect_addr(CLI *c) {
-#ifdef SO_ORIGINAL_DST
-    socklen_t addrlen=sizeof(SOCKADDR_UNION);
-#endif /* SO_ORIGINAL_DST */
-
     /* process "redirect" first */
     if(c->redirect==REDIRECT_ON) {
         s_log(LOG_NOTICE, "Redirecting connection");
@@ -1306,25 +1389,20 @@ NOEXPORT void setup_connect_addr(CLI *c) {
         return;
 
     /* transparent destination */
-#ifdef SO_ORIGINAL_DST
     if(c->opt->option.transparent_dst) {
         c->connect_addr.num=1;
         c->connect_addr.addr=str_alloc(sizeof(SOCKADDR_UNION));
-        if(getsockopt(c->local_rfd.fd, SOL_IP, SO_ORIGINAL_DST,
-                c->connect_addr.addr, &addrlen)) {
-            sockerror("setsockopt SO_ORIGINAL_DST");
+        if(original_dst(c->local_rfd.fd, c->connect_addr.addr))
             longjmp(c->err, 1);
-        }
         return;
     }
-#endif /* SO_ORIGINAL_DST */
 
     /* default "connect" target */
     addrlist_dup(&c->connect_addr, &c->opt->connect_addr);
 }
 
 NOEXPORT void local_bind(CLI *c) {
-#ifndef USE_WIN32
+#if defined(__linux__) || (defined(IP_BINDANY) && defined(IPV6_BINDANY))
     int on;
 
     on=1;
@@ -1364,6 +1442,7 @@ NOEXPORT void local_bind(CLI *c) {
     }
 #else
     /* unsupported platform */
+    /* FIXME: move this check to options.c */
     if(c->opt->option.transparent_src) {
         s_log(LOG_ERR, "Transparent proxy in remote mode is not supported"
             " on this platform");
