@@ -1,6 +1,6 @@
 /*
- *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   stunnel       TLS offloading and load-balancing proxy
+ *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -38,6 +38,11 @@
 #include "common.h"
 #include "prototypes.h"
 
+#ifndef OPENSSL_NO_DH
+DH *dh_params=NULL;
+int dh_needed=0;
+#endif /* OPENSSL_NO_DH */
+
 /**************************************** prototypes */
 
 /* SNI */
@@ -46,45 +51,62 @@ NOEXPORT int servername_cb(SSL *, int *, void *);
 NOEXPORT int matches_wildcard(char *, char *);
 #endif
 
-/* DH/ECDH initialization */
+/* DH/ECDH */
 #ifndef OPENSSL_NO_DH
-NOEXPORT int init_dh(SERVICE_OPTIONS *);
-NOEXPORT DH *read_dh(char *);
-NOEXPORT DH *get_dh2048(void);
+NOEXPORT int dh_init(SERVICE_OPTIONS *);
+NOEXPORT DH *dh_read(char *);
 #endif /* OPENSSL_NO_DH */
 #ifndef OPENSSL_NO_ECDH
-NOEXPORT int init_ecdh(SERVICE_OPTIONS *);
+NOEXPORT int ecdh_init(SERVICE_OPTIONS *);
 #endif /* USE_ECDH */
 
-/* loading certificate */
-NOEXPORT int load_cert(SERVICE_OPTIONS *);
-NOEXPORT int load_key_file(SERVICE_OPTIONS *);
-NOEXPORT int load_key_engine(SERVICE_OPTIONS *);
-#if OPENSSL_VERSION_NUMBER>=0x0090700fL
-NOEXPORT int password_cb(char *, int, int, void *);
-#endif
+/* configuration commands */
+NOEXPORT int conf_init(SERVICE_OPTIONS *section);
 
-/* session cache callbacks */
+/* authentication */
+NOEXPORT int auth_init(SERVICE_OPTIONS *);
+#ifndef OPENSSL_NO_PSK
+NOEXPORT unsigned psk_client_callback(SSL *, const char *,
+    char *, unsigned, unsigned char *, unsigned);
+NOEXPORT unsigned psk_server_callback(SSL *, const char *,
+    unsigned char *, unsigned);
+#endif /* !defined(OPENSSL_NO_PSK) */
+NOEXPORT int load_cert_file(SERVICE_OPTIONS *);
+NOEXPORT int load_key_file(SERVICE_OPTIONS *);
+#ifndef OPENSSL_NO_ENGINE
+NOEXPORT int load_cert_engine(SERVICE_OPTIONS *);
+NOEXPORT int load_key_engine(SERVICE_OPTIONS *);
+#endif
+NOEXPORT int password_cb(char *, int, int, void *);
+
+/* session callbacks */
 NOEXPORT int sess_new_cb(SSL *, SSL_SESSION *);
 NOEXPORT SSL_SESSION *sess_get_cb(SSL *, unsigned char *, int, int *);
 NOEXPORT void sess_remove_cb(SSL_CTX *, SSL_SESSION *);
-NOEXPORT void cache_transfer(SSL_CTX *, const unsigned int, const unsigned,
-    const unsigned char *, const unsigned int,
-    const unsigned char *, const unsigned int,
-    unsigned char **, unsigned int *);
+
+/* sessiond interface */
+NOEXPORT void cache_new(SSL *, SSL_SESSION *);
+NOEXPORT SSL_SESSION *cache_get(SSL *, unsigned char *, int);
+NOEXPORT void cache_remove(SSL_CTX *, SSL_SESSION *);
+NOEXPORT void cache_transfer(SSL_CTX *, const u_char, const long,
+    const u_char *, const size_t,
+    const u_char *, const size_t,
+    unsigned char **, size_t *);
 
 /* info callbacks */
-NOEXPORT void info_callback(
-#if OPENSSL_VERSION_NUMBER>=0x0090700fL
-    const
-#endif
-    SSL *, int, int);
+NOEXPORT void info_callback(const SSL *, int, int);
 
 NOEXPORT void sslerror_queue(void);
 NOEXPORT void sslerror_log(unsigned long, char *);
 
 /**************************************** initialize section->ctx */
 
+#if OPENSSL_VERSION_NUMBER<0x10100000L
+typedef long SSL_OPTIONS_TYPE;
+#else /* OpenSSL >= 1.1.0 */
+typedef long unsigned SSL_OPTIONS_TYPE;
+#endif
+
 int context_init(SERVICE_OPTIONS *section) { /* init SSL context */
     /* create SSL context */
     if(section->option.client)
@@ -95,10 +117,10 @@ int context_init(SERVICE_OPTIONS *section) { /* init SSL context */
         sslerror("SSL_CTX_new");
         return 1; /* FAILED */
     }
-    SSL_CTX_set_ex_data(section->ctx, opt_index, section); /* for callbacks */
+    SSL_CTX_set_ex_data(section->ctx, index_opt, section); /* for callbacks */
 
     /* load certificate and private key to be verified by the peer server */
-#if defined(HAVE_OSSL_ENGINE_H) && OPENSSL_VERSION_NUMBER>=0x0090809fL
+#if !defined(OPENSSL_NO_ENGINE) && OPENSSL_VERSION_NUMBER>=0x0090809fL
     /* SSL_CTX_set_client_cert_engine() was introduced in OpenSSL 0.9.8i */
     if(section->option.client && section->engine) {
         if(SSL_CTX_set_client_cert_engine(section->ctx, section->engine))
@@ -108,7 +130,7 @@ int context_init(SERVICE_OPTIONS *section) { /* init SSL context */
             sslerror("SSL_CTX_set_client_cert_engine"); /* ignore error */
     }
 #endif
-    if(load_cert(section))
+    if(auth_init(section))
         return 1; /* FAILED */
 
     /* initialize verification of the peer server certificate */
@@ -122,16 +144,16 @@ int context_init(SERVICE_OPTIONS *section) { /* init SSL context */
         SSL_CTX_set_tlsext_servername_callback(section->ctx, servername_cb);
 #endif /* OPENSSL_NO_TLSEXT */
 #ifndef OPENSSL_NO_DH
-        init_dh(section); /* ignore the result (errors are not critical) */
+        dh_init(section); /* ignore the result (errors are not critical) */
 #endif /* OPENSSL_NO_DH */
 #ifndef OPENSSL_NO_ECDH
-        init_ecdh(section); /* ignore the result (errors are not critical) */
+        ecdh_init(section); /* ignore the result (errors are not critical) */
 #endif /* OPENSSL_NO_ECDH */
     }
 
     /* setup session cache */
     if(!section->option.client) {
-        unsigned int servname_len=strlen(section->servname);
+        unsigned servname_len=(unsigned)strlen(section->servname);
         if(servname_len>SSL_MAX_SSL_SESSION_ID_LENGTH)
             servname_len=SSL_MAX_SSL_SESSION_ID_LENGTH;
         if(!SSL_CTX_set_session_id_context(section->ctx,
@@ -140,14 +162,16 @@ int context_init(SERVICE_OPTIONS *section) { /* init SSL context */
             return 1; /* FAILED */
         }
     }
-    SSL_CTX_set_session_cache_mode(section->ctx, SSL_SESS_CACHE_BOTH);
+#ifdef SSL_SESS_CACHE_NO_INTERNAL_STORE
+    /* the default cache mode is just SSL_SESS_CACHE_SERVER */
+    SSL_CTX_set_session_cache_mode(section->ctx,
+        SSL_SESS_CACHE_SERVER|SSL_SESS_CACHE_NO_INTERNAL_STORE);
+#endif
     SSL_CTX_sess_set_cache_size(section->ctx, section->session_size);
     SSL_CTX_set_timeout(section->ctx, section->session_timeout);
-    if(section->option.sessiond) {
-        SSL_CTX_sess_set_new_cb(section->ctx, sess_new_cb);
-        SSL_CTX_sess_set_get_cb(section->ctx, sess_get_cb);
-        SSL_CTX_sess_set_remove_cb(section->ctx, sess_remove_cb);
-    }
+    SSL_CTX_sess_set_new_cb(section->ctx, sess_new_cb);
+    SSL_CTX_sess_set_get_cb(section->ctx, sess_get_cb);
+    SSL_CTX_sess_set_remove_cb(section->ctx, sess_remove_cb);
 
     /* set info callback */
     SSL_CTX_set_info_callback(section->ctx, info_callback);
@@ -158,11 +182,24 @@ int context_init(SERVICE_OPTIONS *section) { /* init SSL context */
             sslerror("SSL_CTX_set_cipher_list");
             return 1; /* FAILED */
         }
-    SSL_CTX_set_options(section->ctx, section->ssl_options_set);
-    SSL_CTX_clear_options(section->ctx, section->ssl_options_clear);
+    SSL_CTX_set_options(section->ctx,
+        (SSL_OPTIONS_TYPE)(section->ssl_options_set));
+#if OPENSSL_VERSION_NUMBER>=0x009080dfL
+    SSL_CTX_clear_options(section->ctx,
+        (SSL_OPTIONS_TYPE)(section->ssl_options_clear));
     s_log(LOG_DEBUG, "SSL options: 0x%08lX (+0x%08lX, -0x%08lX)",
         SSL_CTX_get_options(section->ctx),
         section->ssl_options_set, section->ssl_options_clear);
+#else /* OpenSSL older than 0.9.8m */
+    s_log(LOG_DEBUG, "SSL options: 0x%08lX (+0x%08lX)",
+        SSL_CTX_get_options(section->ctx),
+        section->ssl_options_set);
+#endif /* OpenSSL 0.9.8m or later */
+
+    /* initialize OpenSSL CONF options */
+    if(conf_init(section))
+        return 1; /* FAILED */
+
 #ifdef SSL_MODE_RELEASE_BUFFERS
     SSL_CTX_set_mode(section->ctx,
         SSL_MODE_ENABLE_PARTIAL_WRITE |
@@ -190,7 +227,7 @@ NOEXPORT int servername_cb(SSL *ssl, int *ad, void *arg) {
 #endif /* USE_LIBWRAP */
 
     /* leave the alert type at SSL_AD_UNRECOGNIZED_NAME */
-    (void)ad; /* skip warning about unused parameter */
+    (void)ad; /* squash the unused parameter warning */
     if(!section->servername_list_head) {
         s_log(LOG_DEBUG, "SNI: no virtual services defined");
         return SSL_TLSEXT_ERR_OK;
@@ -204,7 +241,7 @@ NOEXPORT int servername_cb(SSL *ssl, int *ad, void *arg) {
     for(list=section->servername_list_head; list; list=list->next)
         if(matches_wildcard((char *)servername, list->servername)) {
             s_log(LOG_DEBUG, "SNI: matched pattern: %s", list->servername);
-            c=SSL_get_ex_data(ssl, cli_index);
+            c=SSL_get_ex_data(ssl, index_cli);
             c->opt=list->opt;
             SSL_set_SSL_CTX(ssl, c->opt->ctx);
             SSL_set_verify(ssl, SSL_CTX_get_verify_mode(c->opt->ctx),
@@ -228,12 +265,12 @@ NOEXPORT int servername_cb(SSL *ssl, int *ad, void *arg) {
  *  - SSL_TLSEXT_ERR_NOACK */
 
 NOEXPORT int matches_wildcard(char *servername, char *pattern) {
-    int diff;
+    ssize_t diff;
 
     if(!servername || !pattern)
         return 0;
     if(*pattern=='*') { /* wildcard comparison */
-        diff=strlen(servername)-strlen(++pattern);
+        diff=(ssize_t)strlen(servername)-(ssize_t)strlen(++pattern);
         if(diff<0) /* pattern longer than servername */
             return 0;
         servername+=diff;
@@ -247,27 +284,30 @@ NOEXPORT int matches_wildcard(char *servername, char *pattern) {
 
 #ifndef OPENSSL_NO_DH
 
-NOEXPORT int init_dh(SERVICE_OPTIONS *section) {
+NOEXPORT int dh_init(SERVICE_OPTIONS *section) {
     DH *dh=NULL;
 
     s_log(LOG_DEBUG, "DH initialization");
-#ifdef HAVE_OSSL_ENGINE_H
+#ifndef OPENSSL_NO_ENGINE
     if(!section->engine) /* cert is a file and not an identifier */
 #endif
-        dh=read_dh(section->cert);
-    if(!dh)
-        dh=get_dh2048();
-    if(!dh) {
-        s_log(LOG_NOTICE, "DH initialization failed");
-        return 1; /* FAILED */
+        dh=dh_read(section->cert);
+    if(dh) {
+        SSL_CTX_set_tmp_dh(section->ctx, dh);
+        s_log(LOG_INFO, "%d-bit DH parameters loaded", 8*DH_size(dh));
+        DH_free(dh);
+        return 0; /* OK */
     }
-    SSL_CTX_set_tmp_dh(section->ctx, dh);
-    s_log(LOG_DEBUG, "DH initialized with %d-bit key", 8*DH_size(dh));
-    DH_free(dh);
+    CRYPTO_r_lock(stunnel_locks[LOCK_DH]);
+    SSL_CTX_set_tmp_dh(section->ctx, dh_params);
+    CRYPTO_r_unlock(stunnel_locks[LOCK_DH]);
+    dh_needed=1; /* generate temporary DH parameters in cron */
+    section->option.dh_needed=1; /* update this context */
+    s_log(LOG_INFO, "Using dynamic DH parameters");
     return 0; /* OK */
 }
 
-NOEXPORT DH *read_dh(char *cert) {
+NOEXPORT DH *dh_read(char *cert) {
     DH *dh;
     BIO *bio;
 
@@ -292,52 +332,12 @@ NOEXPORT DH *read_dh(char *cert) {
     return dh;
 }
 
-NOEXPORT DH *get_dh2048() {
-    static unsigned char dh2048_p[]={ /* OpenSSL DH parameters */
-        0xED,0x92,0x89,0x35,0x82,0x45,0x55,0xCB,0x3B,0xFB,0xA2,0x76,
-        0x5A,0x69,0x04,0x61,0xBF,0x21,0xF3,0xAB,0x53,0xD2,0xCD,0x21,
-        0xDA,0xFF,0x78,0x19,0x11,0x52,0xF1,0x0E,0xC1,0xE2,0x55,0xBD,
-        0x68,0x6F,0x68,0x00,0x53,0xB9,0x22,0x6A,0x2F,0xE4,0x9A,0x34,
-        0x1F,0x65,0xCC,0x59,0x32,0x8A,0xBD,0xB1,0xDB,0x49,0xED,0xDF,
-        0xA7,0x12,0x66,0xC3,0xFD,0x21,0x04,0x70,0x18,0xF0,0x7F,0xD6,
-        0xF7,0x58,0x51,0x19,0x72,0x82,0x7B,0x22,0xA9,0x34,0x18,0x1D,
-        0x2F,0xCB,0x21,0xCF,0x6D,0x92,0xAE,0x43,0xB6,0xA8,0x29,0xC7,
-        0x27,0xA3,0xCB,0x00,0xC5,0xF2,0xE5,0xFB,0x0A,0xA4,0x59,0x85,
-        0xA2,0xBD,0xAD,0x45,0xF0,0xB3,0xAD,0xF9,0xE0,0x81,0x35,0xEE,
-        0xD9,0x83,0xB3,0xCC,0xAE,0xEA,0xEB,0x66,0xE6,0xA9,0x57,0x66,
-        0xB9,0xF1,0x28,0xA5,0x3F,0x22,0x80,0xD7,0x0B,0xA6,0xF6,0x71,
-        0x93,0x9B,0x81,0x0E,0xF8,0x5A,0x90,0xE6,0xCC,0xCA,0x6F,0x66,
-        0x5F,0x7A,0xC0,0x10,0x1A,0x1E,0xF0,0xFC,0x2D,0xB6,0x08,0x0C,
-        0x62,0x28,0xB0,0xEC,0xDB,0x89,0x28,0xEE,0x0C,0xA8,0x3D,0x65,
-        0x94,0x69,0x16,0x69,0x53,0x3C,0x53,0x60,0x13,0xB0,0x2B,0xA7,
-        0xD4,0x82,0x87,0xAD,0x1C,0x72,0x9E,0x41,0x35,0xFC,0xC2,0x7C,
-        0xE9,0x51,0xDE,0x61,0x85,0xFC,0x19,0x9B,0x76,0x60,0x0F,0x33,
-        0xF8,0x6B,0xB3,0xCA,0x52,0x0E,0x29,0xC3,0x07,0xE8,0x90,0x16,
-        0xCC,0xCC,0x00,0x19,0xB6,0xAD,0xC3,0xA4,0x30,0x8B,0x33,0xA1,
-        0xAF,0xD8,0x8C,0x8D,0x9D,0x01,0xDB,0xA4,0xC4,0xDD,0x7F,0x0B,
-        0xBD,0x6F,0x38,0xC3,};
-    static unsigned char dh2048_g[]={0x02,};
-    DH *dh;
-
-    dh=DH_new();
-    if(!dh)
-        return NULL;
-    dh->p=BN_bin2bn(dh2048_p, sizeof dh2048_p, NULL);
-    dh->g=BN_bin2bn(dh2048_g, sizeof dh2048_g, NULL);
-    if(!dh->p || !dh->g) {
-        DH_free(dh);
-        return NULL;
-    }
-    s_log(LOG_DEBUG, "Using hardcoded DH parameters");
-    return dh;
-}
-
 #endif /* OPENSSL_NO_DH */
 
 /**************************************** ECDH initialization */
 
 #ifndef OPENSSL_NO_ECDH
-NOEXPORT int init_ecdh(SERVICE_OPTIONS *section) {
+NOEXPORT int ecdh_init(SERVICE_OPTIONS *section) {
     EC_KEY *ecdh;
 
     s_log(LOG_DEBUG, "ECDH initialization");
@@ -356,39 +356,103 @@ NOEXPORT int init_ecdh(SERVICE_OPTIONS *section) {
 }
 #endif /* OPENSSL_NO_ECDH */
 
-/**************************************** loading certificate */
+/**************************************** initialize OpenSSL CONF */
 
-static int cache_initialized=0;
+NOEXPORT int conf_init(SERVICE_OPTIONS *section) {
+#if OPENSSL_VERSION_NUMBER>=0x10002000L
+    SSL_CONF_CTX *cctx;
+    NAME_LIST *curr;
+    char *cmd, *param;
 
-NOEXPORT int load_cert(SERVICE_OPTIONS *section) {
-    /* load the certificate */
-    if(section->cert) {
-        s_log(LOG_INFO, "Loading cert from file: %s", section->cert);
-        if(!SSL_CTX_use_certificate_chain_file(section->ctx, section->cert)) {
-            sslerror("SSL_CTX_use_certificate_chain_file");
+    if(!section->config)
+        return 0; /* OK */
+    cctx=SSL_CONF_CTX_new();
+    if(!cctx) {
+        sslerror("SSL_CONF_CTX_new");
+        return 1; /* FAILED */
+    }
+    SSL_CONF_CTX_set_ssl_ctx(cctx, section->ctx);
+    SSL_CONF_CTX_set_flags(cctx, SSL_CONF_FLAG_FILE);
+    SSL_CONF_CTX_set_flags(cctx, section->option.client ?
+        SSL_CONF_FLAG_CLIENT : SSL_CONF_FLAG_SERVER);
+    SSL_CONF_CTX_set_flags(cctx, SSL_CONF_FLAG_CERTIFICATE);
+
+    for(curr=section->config; curr; curr=curr->next) {
+        cmd=str_dup(curr->name);
+        param=strchr(cmd, ':');
+        if(param)
+            *param++='\0';
+        switch(SSL_CONF_cmd(cctx, cmd, param)) {
+        case 2:
+            s_log(LOG_DEBUG, "OpenSSL config \"%s\" set to \"%s\"", cmd, param);
+            break;
+        case 1:
+            s_log(LOG_DEBUG, "OpenSSL config command \"%s\" executed", cmd);
+            break;
+        case -2:
+            s_log(LOG_ERR,
+                "OpenSSL config command \"%s\" was not recognised", cmd);
+            str_free(cmd);
+            SSL_CONF_CTX_free(cctx);
+            return 1; /* FAILED */
+        case -3:
+            s_log(LOG_ERR,
+                "OpenSSL config command \"%s\" requires a parameter", cmd);
+            str_free(cmd);
+            SSL_CONF_CTX_free(cctx);
+            return 1; /* FAILED */
+        default:
+            sslerror("SSL_CONF_cmd");
+            str_free(cmd);
+            SSL_CONF_CTX_free(cctx);
             return 1; /* FAILED */
         }
+        str_free(cmd);
     }
 
-    /* load the private key */
-    if(!section->key)
-        section->key=section->cert;
-    if(!section->key) {
-        s_log(LOG_DEBUG, "No private key specified");
+    if(!SSL_CONF_CTX_finish(cctx)) {
+        sslerror("SSL_CONF_CTX_finish");
+        SSL_CONF_CTX_free(cctx);
+        return 1; /* FAILED */
+    }
+    SSL_CONF_CTX_free(cctx);
+#else /* OpenSSL earlier than 1.0.2 */
+    (void)section; /* squash the unused parameter warning */
+#endif /* OpenSSL 1.0.2 or later */
+    return 0; /* OK */
+}
+
+/**************************************** initialize authentication */
+
+NOEXPORT int auth_init(SERVICE_OPTIONS *section) {
+    int cert_needed=1, key_needed=1;
+
+#ifndef OPENSSL_NO_PSK
+    if(section->psk_keys) {
+        if(section->option.client)
+            SSL_CTX_set_psk_client_callback(section->ctx, psk_client_callback);
+        else
+            SSL_CTX_set_psk_server_callback(section->ctx, psk_server_callback);
+    }
+#endif /* !defined(OPENSSL_NO_PSK) */
+
+    /* load the certificate and private key */
+    if(!section->cert || !section->key) {
+        s_log(LOG_DEBUG, "No certificate or private key specified");
         return 0; /* OK */
     }
-#ifdef HAVE_OSSL_ENGINE_H
-    if(section->engine) {
-        if(load_key_engine(section))
-            return 1; /* FAILED */
-    } else
-#endif
-    {
-        if(load_key_file(section))
-            return 1; /* FAILED */
+#ifndef OPENSSL_NO_ENGINE
+    if(section->engine) { /* try to use the engine first */
+        cert_needed=load_cert_engine(section);
+        key_needed=load_key_engine(section);
     }
+#endif
+    if(cert_needed && load_cert_file(section))
+        return 1; /* FAILED */
+    if(key_needed && load_key_file(section))
+        return 1; /* FAILED */
 
-    /* validate the private key */
+    /* validate the private key against the certificate */
     if(!SSL_CTX_check_private_key(section->ctx)) {
         sslerror("Private key does not match the certificate");
         return 1; /* FAILED */
@@ -397,30 +461,136 @@ NOEXPORT int load_cert(SERVICE_OPTIONS *section) {
     return 0; /* OK */
 }
 
+#ifndef OPENSSL_NO_PSK
+
+NOEXPORT unsigned psk_client_callback(SSL *ssl, const char *hint,
+    char *identity, unsigned max_identity_len,
+    unsigned char *psk, unsigned max_psk_len) {
+    CLI *c;
+    size_t identity_len;
+
+    (void)hint; /* squash the unused parameter warning */
+    c=SSL_get_ex_data(ssl, index_cli);
+    if(!c->opt->psk_selected) {
+        s_log(LOG_ERR, "INTERNAL ERROR: No PSK identity selected");
+        return 0;
+    }
+    /* the source seems to have its buffer large enough for
+     * the trailing null character, but the manual page says
+     * nothing about it -- lets play safe */
+    identity_len=strlen(c->opt->psk_selected->identity)+1;
+    if(identity_len>max_identity_len) {
+        s_log(LOG_ERR, "PSK identity too long (%lu>%d bytes)",
+            (long unsigned)identity_len, max_psk_len);
+        return 0;
+    }
+    if(c->opt->psk_selected->key_len>max_psk_len) {
+        s_log(LOG_ERR, "PSK too long (%lu>%d bytes)",
+            (long unsigned)c->opt->psk_selected->key_len, max_psk_len);
+        return 0;
+    }
+    strcpy(identity, c->opt->psk_selected->identity);
+    memcpy(psk, c->opt->psk_selected->key_val, c->opt->psk_selected->key_len);
+    s_log(LOG_INFO, "PSK client configured for identity \"%s\"", identity);
+    return (unsigned)(c->opt->psk_selected->key_len);
+}
+
+NOEXPORT unsigned psk_server_callback(SSL *ssl, const char *identity,
+    unsigned char *psk, unsigned max_psk_len) {
+    CLI *c;
+    PSK_KEYS *found;
+    size_t len;
+
+    c=SSL_get_ex_data(ssl, index_cli);
+    found=psk_find(&c->opt->psk_sorted, identity);
+    if(found) {
+        len=found->key_len;
+    } else {
+        s_log(LOG_ERR, "No key found for PSK identity \"%s\"", identity);
+        len=0;
+    }
+    if(len>max_psk_len) {
+        s_log(LOG_ERR, "PSK too long (%lu>%d bytes)",
+            (long unsigned)len, max_psk_len);
+        len=0;
+    }
+    if(len) {
+        memcpy(psk, found->key_val, len);
+        s_log(LOG_NOTICE, "Key configured for PSK identity \"%s\"", identity);
+    } else { /* block identity probes if possible */
+        if(max_psk_len>=32 && RAND_bytes(psk, 32)>0) {
+            len=32; /* 256 random bits */
+            s_log(LOG_ERR, "Configured random PSK");
+        } else {
+            s_log(LOG_ERR, "Rejecting with unknown_psk_identity alert");
+        }
+    }
+    return (unsigned)len;
+}
+
+NOEXPORT int psk_compar(const void *a, const void *b) {
+    PSK_KEYS *x=*(PSK_KEYS **)a, *y=*(PSK_KEYS **)b;
+
+#if 0
+    s_log(LOG_DEBUG, "PSK cmp: %s %s", x->identity, y->identity);
+#endif
+    return strcmp(x->identity, y->identity);
+}
+
+void psk_sort(PSK_TABLE *table, PSK_KEYS *head) {
+    PSK_KEYS *curr;
+    size_t i;
+
+    table->num=0;
+    for(curr=head; curr; curr=curr->next)
+        ++table->num;
+    s_log(LOG_INFO, "PSK identities: %lu retrieved",
+        (long unsigned)table->num);
+    table->val=str_alloc(table->num*sizeof(PSK_KEYS *));
+    for(curr=head, i=0; i<table->num; ++i) {
+        table->val[i]=curr;
+        curr=curr->next;
+    }
+    qsort(table->val, table->num, sizeof(PSK_KEYS *), psk_compar);
+#if 0
+    for(i=0; i<table->num; ++i)
+        s_log(LOG_DEBUG, "PSK table: %s", table->val[i]->identity);
+#endif
+}
+
+PSK_KEYS *psk_find(const PSK_TABLE *table, const char *identity) {
+    PSK_KEYS key, *ptr=&key, **ret;
+
+    key.identity=(char *)identity;
+    ret=bsearch(&ptr,
+        table->val, table->num, sizeof(PSK_KEYS *), psk_compar);
+    return ret ? *ret : NULL;
+}
+
+#endif /* !defined(OPENSSL_NO_PSK) */
+
+NOEXPORT int load_cert_file(SERVICE_OPTIONS *section) {
+    s_log(LOG_INFO, "Loading certificate from file: %s", section->cert);
+    if(!SSL_CTX_use_certificate_chain_file(section->ctx, section->cert)) {
+        sslerror("SSL_CTX_use_certificate_chain_file");
+        return 1; /* FAILED */
+    }
+    s_log(LOG_INFO, "Certificate loaded from file: %s", section->cert);
+    return 0; /* OK */
+}
+
+static int cache_initialized=0;
+
 NOEXPORT int load_key_file(SERVICE_OPTIONS *section) {
     int i, reason;
     UI_DATA ui_data;
-#if !defined(USE_WIN32) && !defined(USE_OS2)
-    struct stat st; /* buffer for stat */
-#endif
 
-    s_log(LOG_INFO, "Loading key from file: %s", section->key);
-
-#if !defined(USE_WIN32) && !defined(USE_OS2)
-    /* check permissions of the private key file */
-    if(stat(section->key, &st)) {
-        ioerror("Private key file not found");
+    s_log(LOG_INFO, "Loading private key from file: %s", section->key);
+    if(file_permissions(section->key))
         return 1; /* FAILED */
-    }
-    if(st.st_mode & 7)
-        s_log(LOG_WARNING, "Insecure file permissions on %s",
-            section->key);
-#endif
 
     ui_data.section=section; /* setup current section for callbacks */
-#if OPENSSL_VERSION_NUMBER>=0x0090700fL
     SSL_CTX_set_default_passwd_cb(section->ctx, password_cb);
-#endif
 
     for(i=0; i<=3; i++) {
         if(!i && !cache_initialized)
@@ -439,22 +609,44 @@ NOEXPORT int load_key_file(SERVICE_OPTIONS *section) {
         sslerror("SSL_CTX_use_PrivateKey_file");
         return 1; /* FAILED */
     }
+    s_log(LOG_INFO, "Private key loaded from file: %s", section->key);
+    return 0; /* OK */
+}
+
+#ifndef OPENSSL_NO_ENGINE
+
+NOEXPORT int load_cert_engine(SERVICE_OPTIONS *section) {
+    struct {
+        const char *id;
+        X509 *cert;
+    } parms;
+
+    s_log(LOG_INFO, "Loading certificate from engine ID: %s", section->cert);
+    parms.id=section->cert;
+    parms.cert=NULL;
+    ENGINE_ctrl_cmd(section->engine, "LOAD_CERT_CTRL", 0, &parms, NULL, 1);
+    if(!parms.cert) {
+        sslerror("ENGINE_ctrl_cmd");
+        return 1; /* FAILED */
+    }
+    if(!SSL_CTX_use_certificate(section->ctx, parms.cert)) {
+        sslerror("SSL_CTX_use_certificate");
+        return 1; /* FAILED */
+    }
+    s_log(LOG_INFO, "Certificate loaded from engine ID: %s", section->cert);
     return 0; /* OK */
 }
 
-#ifdef HAVE_OSSL_ENGINE_H
 NOEXPORT int load_key_engine(SERVICE_OPTIONS *section) {
     int i, reason;
     UI_DATA ui_data;
     EVP_PKEY *pkey;
     UI_METHOD *ui_method;
 
-    s_log(LOG_INFO, "Loading key from engine: %s", section->key);
+    s_log(LOG_INFO, "Initializing private key on engine ID: %s", section->key);
 
     ui_data.section=section; /* setup current section for callbacks */
-#if OPENSSL_VERSION_NUMBER>=0x0090700fL
     SSL_CTX_set_default_passwd_cb(section->ctx, password_cb);
-#endif
 
 #ifdef USE_WIN32
     ui_method=UI_create_method("stunnel WIN32 UI");
@@ -482,11 +674,12 @@ NOEXPORT int load_key_engine(SERVICE_OPTIONS *section) {
         sslerror("SSL_CTX_use_PrivateKey");
         return 1; /* FAILED */
     }
+    s_log(LOG_INFO, "Private key initialized on engine ID: %s", section->key);
     return 0; /* OK */
 }
-#endif /* HAVE_OSSL_ENGINE_H */
 
-#if OPENSSL_VERSION_NUMBER>=0x0090700fL
+#endif /* !defined(OPENSSL_NO_ENGINE) */
+
 NOEXPORT int password_cb(char *buf, int size, int rwflag, void *userdata) {
     static char cache[PEM_BUFSIZE];
     int len;
@@ -498,21 +691,52 @@ NOEXPORT int password_cb(char *buf, int size, int rwflag, void *userdata) {
 #ifdef USE_WIN32
         len=passwd_cb(buf, size, rwflag, userdata);
 #else
-        /* PEM_def_callback is defined in OpenSSL 0.9.7 and later */
         len=PEM_def_callback(buf, size, rwflag, NULL);
 #endif
-        memcpy(cache, buf, size); /* save in cache */
+        memcpy(cache, buf, (size_t)size); /* save in cache */
         cache_initialized=1;
     } else { /* try the cached value */
-        strncpy(buf, cache, size);
+        strncpy(buf, cache, (size_t)size);
         buf[size-1]='\0';
-        len=strlen(buf);
+        len=(int)strlen(buf);
     }
     return len;
 }
-#endif
 
-/**************************************** session cache callbacks */
+/**************************************** session callbacks */
+
+NOEXPORT int sess_new_cb(SSL *ssl, SSL_SESSION *sess) {
+    CLI *c;
+
+    s_log(LOG_DEBUG, "New session callback");
+    c=SSL_get_ex_data(ssl, index_cli);
+    if(c->opt->option.sessiond)
+        cache_new(ssl, sess);
+    return 1; /* leave the session in local cache for reuse */
+}
+
+NOEXPORT SSL_SESSION *sess_get_cb(SSL *ssl,
+        unsigned char *key, int key_len, int *do_copy) {
+    CLI *c;
+
+    s_log(LOG_DEBUG, "Get session callback");
+    *do_copy=0; /* allow the session to be freed automatically */
+    c=SSL_get_ex_data(ssl, index_cli);
+    if(c->opt->option.sessiond)
+        return cache_get(ssl, key, key_len);
+    return NULL; /* no session to resume */
+}
+
+NOEXPORT void sess_remove_cb(SSL_CTX *ctx, SSL_SESSION *sess) {
+    SERVICE_OPTIONS *opt;
+
+    s_log(LOG_DEBUG, "Remove session callback");
+    opt=SSL_CTX_get_ex_data(ctx, index_opt);
+    if(opt->option.sessiond)
+        cache_remove(ctx, sess);
+}
+
+/**************************************** sessiond functionality */
 
 #define CACHE_CMD_NEW     0x00
 #define CACHE_CMD_GET     0x01
@@ -520,29 +744,36 @@ NOEXPORT int password_cb(char *buf, int size, int rwflag, void *userdata) {
 #define CACHE_RESP_ERR    0x80
 #define CACHE_RESP_OK     0x81
 
-NOEXPORT int sess_new_cb(SSL *ssl, SSL_SESSION *sess) {
+NOEXPORT void cache_new(SSL *ssl, SSL_SESSION *sess) {
     unsigned char *val, *val_tmp;
-    int val_len;
+    ssize_t val_len;
+    const unsigned char *session_id;
+    unsigned int session_id_length;
 
     val_len=i2d_SSL_SESSION(sess, NULL);
-    val_tmp=val=str_alloc(val_len);
+    val_tmp=val=str_alloc((size_t)val_len);
     i2d_SSL_SESSION(sess, &val_tmp);
 
-    cache_transfer(ssl->ctx, CACHE_CMD_NEW, SSL_SESSION_get_timeout(sess),
-        sess->session_id, sess->session_id_length, val, val_len, NULL, NULL);
+#if OPENSSL_VERSION_NUMBER>=0x0090800fL
+    session_id=SSL_SESSION_get_id(sess, &session_id_length);
+#else
+    session_id=(const unsigned char *)sess->session_id;
+    session_id_length=sess->session_id_length;
+#endif
+    cache_transfer(SSL_get_SSL_CTX(ssl), CACHE_CMD_NEW,
+        SSL_SESSION_get_timeout(sess),
+        session_id, session_id_length, val, (size_t)val_len, NULL, NULL);
     str_free(val);
-    return 1; /* leave the session in local cache for reuse */
 }
 
-NOEXPORT SSL_SESSION *sess_get_cb(SSL *ssl,
-        unsigned char *key, int key_len, int *do_copy) {
+NOEXPORT SSL_SESSION *cache_get(SSL *ssl,
+        unsigned char *key, int key_len) {
     unsigned char *val, *val_tmp=NULL;
-    unsigned int val_len=0;
+    ssize_t val_len=0;
     SSL_SESSION *sess;
 
-    *do_copy = 0; /* allow the session to be freed autmatically */
-    cache_transfer(ssl->ctx, CACHE_CMD_GET, 0,
-        key, key_len, NULL, 0, &val, &val_len);
+    cache_transfer(SSL_get_SSL_CTX(ssl), CACHE_CMD_GET, 0,
+        key, (size_t)key_len, NULL, 0, &val, (size_t *)&val_len);
     if(!val)
         return NULL;
     val_tmp=val;
@@ -550,14 +781,23 @@ NOEXPORT SSL_SESSION *sess_get_cb(SSL *ssl,
 #if OPENSSL_VERSION_NUMBER>=0x0090800fL
         (const unsigned char **)
 #endif /* OpenSSL version >= 0.8.0 */
-        &val_tmp, val_len);
+        &val_tmp, (long)val_len);
     str_free(val);
     return sess;
 }
 
-NOEXPORT void sess_remove_cb(SSL_CTX *ctx, SSL_SESSION *sess) {
+NOEXPORT void cache_remove(SSL_CTX *ctx, SSL_SESSION *sess) {
+    const unsigned char *session_id;
+    unsigned int session_id_length;
+
+#if OPENSSL_VERSION_NUMBER>=0x0090800fL
+    session_id=SSL_SESSION_get_id(sess, &session_id_length);
+#else
+    session_id=(const unsigned char *)sess->session_id;
+    session_id_length=sess->session_id_length;
+#endif
     cache_transfer(ctx, CACHE_CMD_REMOVE, 0,
-        sess->session_id, sess->session_id_length, NULL, 0, NULL, NULL);
+        session_id, session_id_length, NULL, 0, NULL, NULL);
 }
 
 #define MAX_VAL_LEN 512
@@ -568,16 +808,17 @@ typedef struct {
     u_char val[MAX_VAL_LEN];
 } CACHE_PACKET;
 
-NOEXPORT void cache_transfer(SSL_CTX *ctx, const unsigned int type,
-        const unsigned int timeout,
-        const unsigned char *key, const unsigned int key_len,
-        const unsigned char *val, const unsigned int val_len,
-        unsigned char **ret, unsigned int *ret_len) {
+NOEXPORT void cache_transfer(SSL_CTX *ctx, const u_char type,
+        const long timeout,
+        const u_char *key, const size_t key_len,
+        const u_char *val, const size_t val_len,
+        unsigned char **ret, size_t *ret_len) {
     char session_id_txt[2*SSL_MAX_SSL_SESSION_ID_LENGTH+1];
     const char hex[16]="0123456789ABCDEF";
     const char *type_description[]={"new", "get", "remove"};
-    unsigned int i;
-    int s, len;
+    unsigned i;
+    SOCKET s;
+    ssize_t len;
     struct timeval t;
     CACHE_PACKET *packet;
     SERVICE_OPTIONS *section;
@@ -592,18 +833,18 @@ NOEXPORT void cache_transfer(SSL_CTX *ctx, const unsigned int type,
     }
     session_id_txt[2*i]='\0';
     s_log(LOG_INFO,
-        "cache_transfer: request=%s, timeout=%u, id=%s, length=%d",
-        type_description[type], timeout, session_id_txt, val_len);
+        "cache_transfer: request=%s, timeout=%ld, id=%s, length=%lu",
+        type_description[type], timeout, session_id_txt, (long unsigned)val_len);
 
     /* allocate UDP packet buffer */
     if(key_len>SSL_MAX_SSL_SESSION_ID_LENGTH) {
-        s_log(LOG_ERR, "cache_transfer: session id too big (%d bytes)",
-            key_len);
+        s_log(LOG_ERR, "cache_transfer: session id too big (%lu bytes)",
+            (unsigned long)key_len);
         return;
     }
     if(val_len>MAX_VAL_LEN) {
-        s_log(LOG_ERR, "cache_transfer: encoded session too big (%d bytes)",
-            key_len);
+        s_log(LOG_ERR, "cache_transfer: encoded session too big (%lu bytes)",
+            (unsigned long)key_len);
         return;
     }
     packet=str_alloc(sizeof(CACHE_PACKET));
@@ -617,15 +858,20 @@ NOEXPORT void cache_transfer(SSL_CTX *ctx, const unsigned int type,
 
     /* create the socket */
     s=s_socket(AF_INET, SOCK_DGRAM, 0, 0, "cache_transfer: socket");
-    if(s<0) {
+    if(s==INVALID_SOCKET) {
         str_free(packet);
         return;
     }
 
     /* retrieve pointer to the section structure of this ctx */
-    section=SSL_CTX_get_ex_data(ctx, opt_index);
-    if(sendto(s, (void *)packet, sizeof(CACHE_PACKET)-MAX_VAL_LEN+val_len, 0,
-            &section->sessiond_addr.sa, addr_len(&section->sessiond_addr))<0) {
+    section=SSL_CTX_get_ex_data(ctx, index_opt);
+    if(sendto(s, (void *)packet,
+#ifdef USE_WIN32
+            (int)
+#endif
+            (sizeof(CACHE_PACKET)-MAX_VAL_LEN+val_len),
+            0, &section->sessiond_addr.sa,
+            addr_len(&section->sessiond_addr))<0) {
         sockerror("cache_transfer: sendto");
         closesocket(s);
         str_free(packet);
@@ -674,7 +920,7 @@ NOEXPORT void cache_transfer(SSL_CTX *ctx, const unsigned int type,
         str_free(packet);
         return;
     }
-    *ret_len=len-(sizeof(CACHE_PACKET)-MAX_VAL_LEN);
+    *ret_len=(size_t)len-(sizeof(CACHE_PACKET)-MAX_VAL_LEN);
     *ret=str_alloc(*ret_len);
     s_log(LOG_INFO, "cache_transfer: session found");
     memcpy(*ret, packet->val, *ret_len);
@@ -683,14 +929,12 @@ NOEXPORT void cache_transfer(SSL_CTX *ctx, const unsigned int type,
 
 /**************************************** informational callback */
 
-NOEXPORT void info_callback(
-#if OPENSSL_VERSION_NUMBER>=0x0090700fL
-        const
-#endif
-        SSL *ssl, int where, int ret) {
+NOEXPORT void info_callback(const SSL *ssl, int where, int ret) {
     CLI *c;
+    SSL_CTX *ctx;
+    const char *state_string;
 
-    c=SSL_get_ex_data((SSL *)ssl, cli_index);
+    c=SSL_get_ex_data((SSL *)ssl, index_cli);
     if(c) {
         if((where&SSL_CB_HANDSHAKE_DONE)
                 && c->reneg_state==RENEG_INIT) {
@@ -701,51 +945,67 @@ NOEXPORT void info_callback(
                 && c->reneg_state==RENEG_ESTABLISHED) {
             int state=SSL_get_state((SSL *)ssl);
 
+#ifndef SSL3_ST_SR_CLNT_HELLO_A
+            if(state==TLS_ST_SR_CLNT_HELLO
+                    || state==TLS_ST_SR_CLNT_HELLO) {
+#else
             if(state==SSL3_ST_SR_CLNT_HELLO_A
                     || state==SSL23_ST_SR_CLNT_HELLO_A) {
+#endif
                 /* client hello received after initial handshake,
                  * this means renegotiation -> mark it */
                 c->reneg_state=RENEG_DETECTED;
             }
         }
+        if(c->opt->log_level<LOG_DEBUG) /* performance optimization */
+            return;
     }
 
-    if(global_options.debug_level<LOG_DEBUG)
-        return; /* performance optimization */
-
     if(where & SSL_CB_LOOP) {
-        s_log(LOG_DEBUG, "SSL state (%s): %s",
-            where & SSL_ST_CONNECT ? "connect" :
-            where & SSL_ST_ACCEPT ? "accept" :
-            "undefined", SSL_state_string_long(ssl));
+        state_string=SSL_state_string_long(ssl);
+        if(strcmp(state_string, "unknown state"))
+            s_log(LOG_DEBUG, "SSL state (%s): %s",
+                where & SSL_ST_CONNECT ? "connect" :
+                where & SSL_ST_ACCEPT ? "accept" :
+                "undefined", state_string);
     } else if(where & SSL_CB_ALERT) {
         s_log(LOG_DEBUG, "SSL alert (%s): %s: %s",
             where & SSL_CB_READ ? "read" : "write",
             SSL_alert_type_string_long(ret),
             SSL_alert_desc_string_long(ret));
     } else if(where==SSL_CB_HANDSHAKE_DONE) {
-        s_log(LOG_DEBUG, "%4ld items in the session cache",
-            SSL_CTX_sess_number(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld client connects (SSL_connect())",
-            SSL_CTX_sess_connect(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld client connects that finished",
-            SSL_CTX_sess_connect_good(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld client renegotiations requested",
-            SSL_CTX_sess_connect_renegotiate(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld server connects (SSL_accept())",
-            SSL_CTX_sess_accept(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld server connects that finished",
-            SSL_CTX_sess_accept_good(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld server renegotiations requested",
-            SSL_CTX_sess_accept_renegotiate(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld session cache hits",
-            SSL_CTX_sess_hits(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld external session cache hits",
-            SSL_CTX_sess_cb_hits(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld session cache misses",
-            SSL_CTX_sess_misses(ssl->ctx));
-        s_log(LOG_DEBUG, "%4ld session cache timeouts",
-            SSL_CTX_sess_timeouts(ssl->ctx));
+        ctx=SSL_get_SSL_CTX((SSL *)ssl);
+        if(c->opt->option.client) {
+            s_log(LOG_DEBUG, "%6ld client connect(s) requested",
+                SSL_CTX_sess_connect(ctx));
+            s_log(LOG_DEBUG, "%6ld client connect(s) succeeded",
+                SSL_CTX_sess_connect_good(ctx));
+            s_log(LOG_DEBUG, "%6ld client renegotiation(s) requested",
+                SSL_CTX_sess_connect_renegotiate(ctx));
+        } else {
+            s_log(LOG_DEBUG, "%6ld server accept(s) requested",
+                SSL_CTX_sess_accept(ctx));
+            s_log(LOG_DEBUG, "%6ld server accept(s) succeeded",
+                SSL_CTX_sess_accept_good(ctx));
+            s_log(LOG_DEBUG, "%6ld server renegotiation(s) requested",
+                SSL_CTX_sess_accept_renegotiate(ctx));
+        }
+        /* according to the source it not only includes internal
+           and external session caches, but also session tickets */
+        s_log(LOG_DEBUG, "%6ld session reuse(s)",
+            SSL_CTX_sess_hits(ctx));
+        if(!c->opt->option.client) { /* server session cache stats */
+            s_log(LOG_DEBUG, "%6ld internal session cache item(s)",
+                SSL_CTX_sess_number(ctx));
+            s_log(LOG_DEBUG, "%6ld internal session cache fill-up(s)",
+                SSL_CTX_sess_cache_full(ctx));
+            s_log(LOG_DEBUG, "%6ld internal session cache miss(es)",
+                SSL_CTX_sess_misses(ctx));
+            s_log(LOG_DEBUG, "%6ld external session cache hit(s)",
+                SSL_CTX_sess_cb_hits(ctx));
+            s_log(LOG_DEBUG, "%6ld expired session(s) retrieved",
+                SSL_CTX_sess_timeouts(ctx));
+        }
     }
 }
 
@@ -776,8 +1036,8 @@ NOEXPORT void sslerror_queue(void) { /* recursive dump of the error queue */
 NOEXPORT void sslerror_log(unsigned long err, char *txt) {
     char *error_string;
 
-    error_string=str_alloc(120);
-    ERR_error_string(err, error_string);
+    error_string=str_alloc(256);
+    ERR_error_string_n(err, error_string, 256);
     s_log(LOG_ERR, "%s: %lX: %s", txt, err, error_string);
     str_free(error_string);
 }
