@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2015 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -75,7 +75,7 @@ void s_poll_add(s_poll_set *fds, SOCKET fd, int rd, int wr) {
 
     for(i=0; i<fds->nfds && fds->ufds[i].fd!=fd; i++)
         ;
-    if(i==fds->nfds) {
+    if(i==fds->nfds) { /* not found */
         if(i==fds->allocated) {
             fds->allocated=i+1;
             s_poll_realloc(fds);
@@ -94,12 +94,24 @@ void s_poll_add(s_poll_set *fds, SOCKET fd, int rd, int wr) {
         fds->ufds[i].events|=POLLOUT;
 }
 
+void s_poll_remove(s_poll_set *fds, SOCKET fd) {
+    unsigned i;
+
+    for(i=0; i<fds->nfds && fds->ufds[i].fd!=fd; i++)
+        ;
+    if(i<fds->nfds) { /* found */
+        memmove(fds->ufds+i, fds->ufds+i+1,
+            (fds->nfds-i-1)*sizeof(struct pollfd));
+        fds->nfds--;
+    }
+}
+
 int s_poll_canread(s_poll_set *fds, SOCKET fd) {
     unsigned i;
 
     for(i=0; i<fds->nfds; i++)
         if(fds->ufds[i].fd==fd)
-            return fds->ufds[i].revents&POLLIN;
+            return fds->ufds[i].revents&(POLLIN|POLLERR);
     return 0; /* not listed in fds */
 }
 
@@ -108,7 +120,7 @@ int s_poll_canwrite(s_poll_set *fds, SOCKET fd) {
 
     for(i=0; i<fds->nfds; i++)
         if(fds->ufds[i].fd==fd)
-            return fds->ufds[i].revents&POLLOUT;
+            return fds->ufds[i].revents&(POLLOUT|POLLERR);
     return 0; /* not listed in fds */
 }
 
@@ -140,6 +152,14 @@ NOEXPORT void s_poll_realloc(s_poll_set *fds) {
     fds->ufds=str_realloc(fds->ufds, fds->allocated*sizeof(struct pollfd));
 }
 
+void s_poll_dump(s_poll_set *fds, int level) {
+    unsigned i;
+
+    for(i=0; i<fds->nfds; i++)
+        s_log(level, "FD=%ld events=0x%X revents=0x%X",
+            (long)fds->ufds[i].fd, fds->ufds[i].events, fds->ufds[i].revents);
+}
+
 #ifdef USE_UCONTEXT
 
 /* move ready contexts from waiting queue to ready queue */
@@ -154,13 +174,14 @@ NOEXPORT void scan_waiting_queue(void) {
 
     time(&now);
     /* count file descriptors */
-    min_timeout=-1;
+    min_timeout=-1; /* infinity */
     nfds=0;
     for(context=waiting_head; context; context=context->next) {
         nfds+=context->fds->nfds;
         if(context->finish>=0) /* finite time */
             if(min_timeout<0 || min_timeout>context->finish-now)
-                min_timeout=context->finish-now<0 ? 0 : context->finish-now;
+                min_timeout=
+                    (int)(context->finish-now<0 ? 0 : context->finish-now);
     }
     /* setup ufds structure */
     if(nfds>max_nfds) { /* need to allocate more memory */
@@ -235,7 +256,7 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
     static CONTEXT *to_free=NULL; /* delayed memory deallocation */
 
     /* FIXME: msec parameter is currently ignored with UCONTEXT threads */
-    (void)msec; /* skip warning about unused parameter */
+    (void)msec; /* squash the unused parameter warning */
 
     /* remove the current context from ready queue */
     context=ready_head;
@@ -349,32 +370,38 @@ void s_poll_add(s_poll_set *fds, SOCKET fd, int rd, int wr) {
     }
 #endif
     if(rd)
-        FD_SET((unsigned)fd, fds->irfds);
+        FD_SET(fd, fds->irfds);
     if(wr)
-        FD_SET((unsigned)fd, fds->iwfds);
+        FD_SET(fd, fds->iwfds);
     /* always expect errors (and the Spanish Inquisition) */
-    FD_SET((unsigned)fd, fds->ixfds);
+    FD_SET(fd, fds->ixfds);
     if(fd>fds->max)
         fds->max=fd;
 }
 
+void s_poll_remove(s_poll_set *fds, SOCKET fd) {
+    FD_CLR(fd, fds->irfds);
+    FD_CLR(fd, fds->iwfds);
+    FD_CLR(fd, fds->ixfds);
+}
+
 int s_poll_canread(s_poll_set *fds, SOCKET fd) {
-    return FD_ISSET(fd, fds->orfds);
+    return FD_ISSET(fd, fds->orfds) || FD_ISSET(fd, fds->oxfds);
 }
 
 int s_poll_canwrite(s_poll_set *fds, SOCKET fd) {
-    return FD_ISSET(fd, fds->owfds);
+    return FD_ISSET(fd, fds->owfds) || FD_ISSET(fd, fds->oxfds);
 }
 
 int s_poll_hup(s_poll_set *fds, SOCKET fd) {
-    (void)fds; /* skip warning about unused parameter */
-    (void)fd; /* skip warning about unused parameter */
+    (void)fds; /* squash the unused parameter warning */
+    (void)fd; /* squash the unused parameter warning */
     return 0; /* FIXME: how to detect HUP condition with select()? */
 }
 
 int s_poll_rdhup(s_poll_set *fds, SOCKET fd) {
-    (void)fds; /* skip warning about unused parameter */
-    (void)fd; /* skip warning about unused parameter */
+    (void)fds; /* squash the unused parameter warning */
+    (void)fd; /* squash the unused parameter warning */
     return 0; /* FIXME: how to detect RDHUP condition with select()? */
 }
 
@@ -414,6 +441,24 @@ NOEXPORT void s_poll_realloc(s_poll_set *fds) {
     fds->oxfds=str_realloc(fds->oxfds, FD_SIZE(fds));
 }
 
+void s_poll_dump(s_poll_set *fds, int level) {
+    SOCKET fd;
+    int ir, iw, ix, or, ow, ox;
+
+    for(fd=0; fd<fds->max; fd++) {
+        ir=FD_ISSET(fd, fds->irfds);
+        iw=FD_ISSET(fd, fds->iwfds);
+        ix=FD_ISSET(fd, fds->ixfds);
+        or=FD_ISSET(fd, fds->orfds);
+        ow=FD_ISSET(fd, fds->owfds);
+        ox=FD_ISSET(fd, fds->oxfds);
+        if(ir || iw || ix || or || ow || ox)
+            s_log(level, "FD=%ld ifds=%c%c%c ofds=%c%c%c", (long)fd,
+                ir?'r':'-', iw?'w':'-', ix?'x':'-',
+                or?'r':'-', ow?'w':'-', ox?'x':'-');
+    }
+}
+
 #endif /* USE_POLL */
 
 /**************************************** fd management */
@@ -482,7 +527,7 @@ int s_connect(CLI *c, SOCKADDR_UNION *addr, socklen_t addrlen) {
     s_log(LOG_INFO, "s_connect: connecting %s", dst);
 
     if(!connect(c->fd, &addr->sa, addrlen)) {
-        s_log(LOG_NOTICE, "s_connect: connected %s", dst);
+        s_log(LOG_INFO, "s_connect: connected %s", dst);
         str_free(dst);
         return 0; /* no error -> success (on some OSes over the loopback) */
     }
@@ -606,9 +651,8 @@ void fd_putline(CLI *c, SOCKET fd, const char *line) {
     tmpline=str_printf("%s%s", line, crlf);
     len=strlen(tmpline);
     s_write(c, fd, tmpline, len);
-    tmpline[len-2]='\0'; /* remove CRLF */
-    s_log(LOG_DEBUG, " -> %s", tmpline);
     str_free(tmpline);
+    s_log(LOG_DEBUG, " -> %s", line);
 }
 
 char *fd_getline(CLI *c, SOCKET fd) {
@@ -731,7 +775,7 @@ char *ssl_getstring(CLI *c) { /* get null-terminated string */
     line=str_alloc(allocated);
     for(;;) {
         if(ptr>65536) { /* >64KB --> DoS protection */
-            s_log(LOG_ERR, "fd_getline: Line too long");
+            s_log(LOG_ERR, "ssl_getstring: Line too long");
             str_free(line);
             longjmp(c->err, 1);
         }
@@ -747,6 +791,54 @@ char *ssl_getstring(CLI *c) { /* get null-terminated string */
     return line;
 }
 
+char *ssl_getline(CLI *c) { /* get newline-terminated string */
+    char *line;
+    size_t ptr=0, allocated=32;
+
+    line=str_alloc(allocated);
+    for(;;) {
+        if(ptr>65536) { /* >64KB --> DoS protection */
+            s_log(LOG_ERR, "ssl_getline: Line too long");
+            str_free(line);
+            longjmp(c->err, 1);
+        }
+        if(allocated<ptr+1) {
+            allocated*=2;
+            line=str_realloc(line, allocated);
+        }
+        s_ssl_read(c, line+ptr, 1);
+        if(line[ptr]=='\r')
+            continue;
+        if(line[ptr]=='\n')
+            break;
+        if(line[ptr]=='\0')
+            break;
+        ++ptr;
+    }
+    line[ptr]='\0';
+    s_log(LOG_DEBUG, " <- %s", line);
+    return line;
+}
+
+void ssl_putline(CLI *c, const char *line) { /* put newline-terminated string */
+    char *tmpline;
+    const char crlf[]="\r\n";
+    size_t len;
+
+    tmpline=str_printf("%s%s", line, crlf);
+    len=strlen(tmpline);
+    if(len>INT_MAX) { /* paranoia */
+        s_log(LOG_ERR, "ssl_putline: Line too long");
+        str_free(tmpline);
+        longjmp(c->err, 1);
+    }
+    s_ssl_write(c, tmpline, (int)len);
+    str_free(tmpline);
+    s_log(LOG_DEBUG, " -> %s", line);
+}
+
+/**************************************** network helpers */
+
 #define INET_SOCKET_PAIR
 
 int make_sockets(SOCKET fd[2]) { /* make a pair of connected ipv4 sockets */
@@ -810,4 +902,26 @@ int make_sockets(SOCKET fd[2]) { /* make a pair of connected ipv4 sockets */
     return 0;
 }
 
+/* returns 0 on success, and -1 on error */
+int original_dst(const SOCKET fd, SOCKADDR_UNION *addr) {
+    socklen_t addrlen;
+
+    memset(addr, 0, sizeof(SOCKADDR_UNION));
+    addrlen=sizeof(SOCKADDR_UNION);
+#ifdef SO_ORIGINAL_DST
+#ifdef USE_IPv6
+    if(!getsockopt(fd, SOL_IPV6, SO_ORIGINAL_DST, &addr->sa, &addrlen))
+        return 0; /* succeeded */
+#endif /* USE_IPv6 */
+    if(!getsockopt(fd, SOL_IP, SO_ORIGINAL_DST, &addr->sa, &addrlen))
+        return 0; /* succeeded */
+    sockerror("getsockopt SO_ORIGINAL_DST");
+#else /* SO_ORIGINAL_DST */
+    if(!getsockname(fd, &addr->sa, &addrlen))
+        return 0; /* succeeded */
+    sockerror("getsockname");
+#endif /* SO_ORIGINAL_DST */
+    return -1; /* failed */
+}
+
 /* end of network.c */
