@@ -1,6 +1,6 @@
 /*
- *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   stunnel       TLS offloading and load-balancing proxy
+ *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -40,8 +40,11 @@
 
 /**************************************** prototypes */
 
-NOEXPORT int name2addrlist(SOCKADDR_LIST *, char *, char *);
-NOEXPORT int hostport2addrlist(SOCKADDR_LIST *, char *, char *);
+#if defined(USE_WIN32) && !defined(_WIN32_WCE)
+NOEXPORT int get_ipv6(LPTSTR);
+#endif
+NOEXPORT void addrlist2addr(SOCKADDR_UNION *, SOCKADDR_LIST *);
+NOEXPORT void addrlist_reset(SOCKADDR_LIST *);
 
 #ifndef HAVE_GETADDRINFO
 
@@ -72,6 +75,10 @@ struct addrinfo {
 };
 #endif
 
+#ifndef AI_PASSIVE
+#define AI_PASSIVE 1
+#endif
+
 NOEXPORT int getaddrinfo(const char *, const char *,
     const struct addrinfo *, struct addrinfo **);
 NOEXPORT int alloc_addresses(struct hostent *, const struct addrinfo *,
@@ -90,83 +97,90 @@ GETNAMEINFO s_getnameinfo;
 
 void resolver_init() {
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
-    HINSTANCE handle;
-
-        /* IPv6 in Windows XP or higher */
-    handle=LoadLibrary(TEXT("ws2_32.dll"));
-    if(handle) {
-        s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
-        s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
-        s_getnameinfo=(GETNAMEINFO)GetProcAddress(handle, "getnameinfo");
-        if(s_getaddrinfo && s_freeaddrinfo && s_getnameinfo)
-            return; /* IPv6 detected -> OK */
-        FreeLibrary(handle);
-    }
-
-        /* experimental IPv6 for Windows 2000 */
-    handle=LoadLibrary(TEXT("wship6.dll"));
-    if(handle) {
-        s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
-        s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
-        s_getnameinfo=(GETNAMEINFO)GetProcAddress(handle, "getnameinfo");
-        if(s_getaddrinfo && s_freeaddrinfo && s_getnameinfo)
-            return; /* IPv6 detected -> OK */
-        FreeLibrary(handle);
-    }
-
-        /* fall back to the built-in emulation */
-    s_getaddrinfo=NULL;
-    s_freeaddrinfo=NULL;
-    s_getnameinfo=NULL;
+    if(get_ipv6(TEXT("ws2_32.dll"))) /* IPv6 in Windows XP or higher */
+        return;
+    if(get_ipv6(TEXT("wship6.dll"))) /* experimental IPv6 for Windows 2000 */
+        return;
+    /* fall back to the built-in emulation */
 #endif
 }
 
+#if defined(USE_WIN32) && !defined(_WIN32_WCE)
+NOEXPORT int get_ipv6(LPTSTR file) {
+    HINSTANCE handle;
+
+    handle=LoadLibrary(file);
+    if(!handle)
+        return 0;
+    s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
+    s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
+    s_getnameinfo=(GETNAMEINFO)GetProcAddress(handle, "getnameinfo");
+    if(!s_getaddrinfo || !s_freeaddrinfo || !s_getnameinfo) {
+        s_getaddrinfo=NULL;
+        s_freeaddrinfo=NULL;
+        s_getnameinfo=NULL;
+        FreeLibrary(handle);
+        return 0;
+    }
+    return 1; /* IPv6 detected -> OK */
+}
+#endif
+
 /**************************************** stunnel resolver API */
 
-int name2addr(SOCKADDR_UNION *addr, char *name, char *default_host) {
-    SOCKADDR_LIST addr_list;
-    int retval;
+unsigned name2addr(SOCKADDR_UNION *addr, char *name, int passive) {
+    SOCKADDR_LIST *addr_list;
+    unsigned retval;
 
-    addr_list.num=0;
-    addr_list.addr=NULL;
-    retval=name2addrlist(&addr_list, name, default_host);
-    if(retval>0)
-        memcpy(addr, &addr_list.addr[0], sizeof *addr);
-    if(addr_list.addr)
-        str_free(addr_list.addr);
+    addr_list=str_alloc(sizeof(SOCKADDR_LIST));
+    addrlist_clear(addr_list, passive);
+    retval=name2addrlist(addr_list, name);
+    if(retval)
+        addrlist2addr(addr, addr_list);
+    str_free(addr_list->addr);
+    str_free(addr_list);
     return retval;
 }
 
-int hostport2addr(SOCKADDR_UNION *addr, char *hostname, char *portname) {
-    SOCKADDR_LIST addr_list;
-    int retval;
+unsigned hostport2addr(SOCKADDR_UNION *addr,
+        char *host_name, char *port_name, int passive) {
+    SOCKADDR_LIST *addr_list;
+    unsigned retval;
 
-    addr_list.num=0;
-    addr_list.addr=NULL;
-    retval=hostport2addrlist(&addr_list, hostname, portname);
-    if(retval>0)
-        memcpy(addr, &addr_list.addr[0], sizeof *addr);
-    if(addr_list.addr)
-        str_free(addr_list.addr);
+    addr_list=str_alloc(sizeof(SOCKADDR_LIST));
+    addrlist_clear(addr_list, passive);
+    retval=hostport2addrlist(addr_list, host_name, port_name);
+    if(retval)
+        addrlist2addr(addr, addr_list);
+    str_free(addr_list->addr);
+    str_free(addr_list);
     return retval;
 }
 
-int namelist2addrlist(SOCKADDR_LIST *addr_list, NAME_LIST *name_list, char *default_host) {
-    /* recursive implementation to reverse the list */
-    if(!name_list)
-        return 0;
-    return namelist2addrlist(addr_list, name_list->next, default_host) +
-        name2addrlist(addr_list, name_list->name, default_host);
+NOEXPORT void addrlist2addr(SOCKADDR_UNION *addr, SOCKADDR_LIST *addr_list) {
+    unsigned i;
+
+    for(i=0; i<addr_list->num; ++i) { /* find the first IPv4 address */
+        if(addr_list->addr[i].in.sin_family==AF_INET) {
+            memcpy(addr, &addr_list->addr[i], sizeof(SOCKADDR_UNION));
+            return;
+        }
+    }
+#ifdef USE_IPv6
+    for(i=0; i<addr_list->num; ++i) { /* find the first IPv6 address */
+        if(addr_list->addr[i].in.sin_family==AF_INET6) {
+            memcpy(addr, &addr_list->addr[i], sizeof(SOCKADDR_UNION));
+            return;
+        }
+    }
+#endif
+    /* copy the first address resolved (curently AF_UNIX) */
+    memcpy(addr, &addr_list->addr[0], sizeof(SOCKADDR_UNION));
 }
 
-NOEXPORT int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_host) {
-    char *tmp, *hostname, *portname;
-    int retval;
-
-    addr_list->rr_val=0; /* reset round-robin counter */
-    /* allow structures created with sockaddr_dup() to modify
-     * the original rr_val rather than its local copy */
-    addr_list->rr_ptr=&addr_list->rr_val;
+unsigned name2addrlist(SOCKADDR_LIST *addr_list, char *name) {
+    char *tmp, *host_name, *port_name;
+    unsigned retval;
 
     /* first check if this is a UNIX socket */
 #ifdef HAVE_STRUCT_SOCKADDR_UN
@@ -184,54 +198,69 @@ NOEXPORT int name2addrlist(SOCKADDR_LIST *addr_list, char *name, char *default_h
     }
 #endif
 
-    /* set hostname and portname */
+    /* setup host_name and port_name */
     tmp=str_dup(name);
-    portname=strrchr(tmp, ':');
-    if(portname) {
-        hostname=tmp;
-        *portname++='\0';
+    port_name=strrchr(tmp, ':');
+    if(port_name) {
+        host_name=tmp;
+        *port_name++='\0';
     } else { /* no ':' - use default host IP */
-        hostname=default_host;
-        portname=tmp;
+        host_name=NULL;
+        port_name=tmp;
     }
 
     /* fill addr_list structure */
-    retval=hostport2addrlist(addr_list, hostname, portname);
+    retval=hostport2addrlist(addr_list, host_name, port_name);
     str_free(tmp);
     return retval;
 }
 
-NOEXPORT int hostport2addrlist(SOCKADDR_LIST *addr_list,
-        char *hostname, char *portname) {
+unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
+        char *host_name, char *port_name) {
     struct addrinfo hints, *res=NULL, *cur;
-    int err, retries=0;
+    int err, retry=0;
 
     memset(&hints, 0, sizeof hints);
 #if defined(USE_IPv6) || defined(USE_WIN32)
-    hints.ai_family=PF_UNSPEC;
+    hints.ai_family=AF_UNSPEC;
 #else
-    hints.ai_family=PF_INET;
+    hints.ai_family=AF_INET;
 #endif
     hints.ai_socktype=SOCK_STREAM;
     hints.ai_protocol=IPPROTO_TCP;
-    for(;;) {
-        err=getaddrinfo(hostname, portname, &hints, &res);
-        if(err && res)
-            freeaddrinfo(res);
-        if(err!=EAI_AGAIN || ++retries>=3)
-            break;
-        s_log(LOG_DEBUG, "getaddrinfo: EAI_AGAIN received: retrying");
-        sleep(1);
+    hints.ai_flags=0;
+    if(addr_list->passive) {
+        hints.ai_family=AF_INET; /* first try IPv4 for passive requests */
+        hints.ai_flags|=AI_PASSIVE;
     }
-    switch(err) {
-    case 0:
-        break; /* success */
-    case EAI_SERVICE:
-        s_log(LOG_ERR, "Unknown TCP service '%s'", portname);
+    for(;;) {
+        err=getaddrinfo(host_name, port_name, &hints, &res);
+        if(!err)
+            break;
+        if(res)
+            freeaddrinfo(res);
+        if(err==EAI_AGAIN && ++retry<=3) {
+            s_log(LOG_DEBUG, "getaddrinfo: EAI_AGAIN received: retrying");
+            sleep(1);
+            continue;
+        }
+#if defined(USE_IPv6) || defined(USE_WIN32)
+        if(hints.ai_family==AF_INET) {
+            hints.ai_family=AF_UNSPEC;
+            continue; /* retry for non-IPv4 addresses */
+        }
+#endif
+        break;
+    }
+    if(err==EAI_SERVICE) {
+        s_log(LOG_ERR, "Unknown TCP service \"%s\"", port_name);
         return 0; /* error */
-    default:
-        s_log(LOG_ERR, "Error resolving '%s': %s",
-            hostname, s_gai_strerror(err));
+    }
+    if(err) {
+        s_log(LOG_ERR, "Error resolving \"%s\": %s",
+            host_name ? host_name :
+                (addr_list->passive ? DEFAULT_ANY : DEFAULT_LOOPBACK),
+            s_gai_strerror(err));
         return 0; /* error */
     }
 
@@ -244,19 +273,61 @@ NOEXPORT int hostport2addrlist(SOCKADDR_LIST *addr_list,
         }
         addr_list->addr=str_realloc(addr_list->addr,
             (addr_list->num+1)*sizeof(SOCKADDR_UNION));
-        memcpy(&addr_list->addr[addr_list->num], cur->ai_addr, cur->ai_addrlen);
+        memcpy(&addr_list->addr[addr_list->num], cur->ai_addr,
+            (size_t)cur->ai_addrlen);
         ++(addr_list->num);
     }
     freeaddrinfo(res);
     return addr_list->num; /* ok - return the number of addresses */
 }
 
-void addrlist_dup(SOCKADDR_LIST *dst, const SOCKADDR_LIST *src) {
+/* initialize the structure */
+void addrlist_clear(SOCKADDR_LIST *addr_list, int passive) {
+    addrlist_reset(addr_list);
+    addr_list->names=NULL;
+    addr_list->passive=passive;
+}
+
+/* prepare the structure to resolve new hosts */
+NOEXPORT void addrlist_reset(SOCKADDR_LIST *addr_list) {
+    addr_list->num=0;
+    addr_list->addr=NULL;
+    addr_list->rr_val=0; /* reset the round-robin counter */
+    /* allow structures created with sockaddr_dup() to modify
+     * the original rr_val rather than its local copy */
+    addr_list->rr_ptr=&addr_list->rr_val;
+}
+
+unsigned addrlist_dup(SOCKADDR_LIST *dst, const SOCKADDR_LIST *src) {
     memcpy(dst, src, sizeof(SOCKADDR_LIST));
-    if(src->addr) {
+    if(src->num) { /* already resolved */
         dst->addr=str_alloc(src->num*sizeof(SOCKADDR_UNION));
         memcpy(dst->addr, src->addr, src->num*sizeof(SOCKADDR_UNION));
+    } else { /* delayed resolver */
+        addrlist_resolve(dst);
     }
+    return dst->num;
+}
+
+unsigned addrlist_resolve(SOCKADDR_LIST *addr_list) {
+    unsigned num=0, rnd;
+    NAME_LIST *host;
+
+    addrlist_reset(addr_list);
+    for(host=addr_list->names; host; host=host->next)
+        num+=name2addrlist(addr_list, host->name);
+    switch(num) {
+    case 0:
+    case 1:
+        addr_list->rr_val=0;
+        break;
+    default:
+        /* randomize the initial value of round-robin counter */
+        /* ignore the error value and the distribution bias */
+        RAND_bytes((unsigned char *)&rnd, sizeof rnd);
+        addr_list->rr_val=rnd%num;
+    }
+    return num;
 }
 
 char *s_ntop(SOCKADDR_UNION *addr, socklen_t addrlen) {
@@ -310,6 +381,8 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
     int retval;
     char *tmpstr;
 
+    if(!node)
+        node=hints->ai_flags&AI_PASSIVE ? DEFAULT_ANY : DEFAULT_LOOPBACK;
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
     if(s_getaddrinfo)
         return s_getaddrinfo(node, service, hints, res);
@@ -323,7 +396,7 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
         p=getservbyname(service, "tcp");
         if(!p)
             return EAI_NONAME;
-        port=p->s_port;
+        port=(u_short)p->s_port;
 #endif /* defined(_WIN32_WCE) */
     }
 
@@ -336,7 +409,7 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
 #if defined(USE_IPv6) && !defined(USE_WIN32)
     ai->ai_family=AF_INET6;
     ai->ai_addrlen=sizeof(struct sockaddr_in6);
-    ai->ai_addr=str_alloc(ai->ai_addrlen);
+    ai->ai_addr=str_alloc((size_t)ai->ai_addrlen);
     ai->ai_addr->sa_family=AF_INET6;
     if(inet_pton(AF_INET6, node,
             &((struct sockaddr_in6 *)ai->ai_addr)->sin6_addr)>0) {
@@ -359,7 +432,7 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
     /* not numerical: need to call resolver library */
     *res=NULL;
     ai=NULL;
-    enter_critical_section(CRIT_INET);
+    CRYPTO_w_lock(stunnel_locks[LOCK_INET]);
 #ifdef HAVE_GETHOSTBYNAME2
     h=gethostbyname2(node, AF_INET6);
     if(h) /* some IPv6 addresses found */
@@ -377,7 +450,7 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
 #ifdef HAVE_ENDHOSTENT
     endhostent();
 #endif
-    leave_critical_section(CRIT_INET);
+    CRYPTO_w_unlock(stunnel_locks[LOCK_INET]);
     if(retval) { /* error: free allocated memory */
         freeaddrinfo(*res);
         *res=NULL;
@@ -407,18 +480,18 @@ NOEXPORT int alloc_addresses(struct hostent *h, const struct addrinfo *hints,
 #if defined(USE_IPv6)
         if(h->h_addrtype==AF_INET6) {
             ai->ai_addrlen=sizeof(struct sockaddr_in6);
-            ai->ai_addr=str_alloc(ai->ai_addrlen);
+            ai->ai_addr=str_alloc((size_t)ai->ai_addrlen);
             memcpy(&((struct sockaddr_in6 *)ai->ai_addr)->sin6_addr,
-                h->h_addr_list[i], h->h_length);
+                h->h_addr_list[i], (size_t)h->h_length);
         } else
 #endif
         {
             ai->ai_addrlen=sizeof(struct sockaddr_in);
-            ai->ai_addr=str_alloc(ai->ai_addrlen);
+            ai->ai_addr=str_alloc((size_t)ai->ai_addrlen);
             memcpy(&((struct sockaddr_in *)ai->ai_addr)->sin_addr,
-                h->h_addr_list[i], h->h_length);
+                h->h_addr_list[i], (size_t)h->h_length);
         }
-        ai->ai_addr->sa_family=h->h_addrtype;
+        ai->ai_addr->sa_family=(u_short)h->h_addrtype;
         /* offsets of sin_port and sin6_port should be the same */
         ((struct sockaddr_in *)ai->ai_addr)->sin_port=port;
     }
@@ -435,10 +508,8 @@ NOEXPORT void freeaddrinfo(struct addrinfo *current) {
     }
 #endif
     while(current) {
-        if(current->ai_addr)
-            str_free(current->ai_addr);
-        if(current->ai_canonname)
-            str_free(current->ai_canonname);
+        str_free(current->ai_addr);
+        str_free(current->ai_canonname);
         next=current->ai_next;
         str_free(current);
         current=next;
@@ -500,8 +571,8 @@ const char *s_gai_strerror(int err) {
 /* implementation is limited to functionality needed by stunnel */
 
 #ifndef HAVE_GETNAMEINFO
-int getnameinfo(const struct sockaddr *sa, int salen,
-    char *host, int hostlen, char *serv, int servlen, int flags) {
+int getnameinfo(const struct sockaddr *sa, socklen_t salen,
+    char *host, size_t hostlen, char *serv, size_t servlen, int flags) {
 
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
     if(s_getnameinfo)
@@ -514,10 +585,10 @@ int getnameinfo(const struct sockaddr *sa, int salen,
                 (void *)&((struct sockaddr_in *)sa)->sin_addr,
             host, hostlen);
 #else /* USE_IPv6 */
-        enter_critical_section(CRIT_INET); /* inet_ntoa is not mt-safe */
+        CRYPTO_w_lock(stunnel_locks[LOCK_INET]); /* inet_ntoa is not mt-safe */
         strncpy(host, inet_ntoa(((struct sockaddr_in *)sa)->sin_addr),
             hostlen);
-        leave_critical_section(CRIT_INET);
+        CRYPTO_w_unlock(stunnel_locks[LOCK_INET]);
         host[hostlen-1]='\0';
 #endif /* USE_IPv6 */
     }
