@@ -1,6 +1,6 @@
 /*
- *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   stunnel       TLS offloading and load-balancing proxy
+ *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -43,19 +43,11 @@
 #include "common.h"
 #include "prototypes.h"
 
+int stunnel_locks[STUNNEL_LOCKS];
+
 #if defined(USE_UCONTEXT) || defined(USE_FORK)
 /* no need for critical sections */
 
-void enter_critical_section(SECTION_CODE i) {
-    (void)i; /* skip warning about unused parameter */
-    /* empty */
-}
-
-void leave_critical_section(SECTION_CODE i) {
-    (void)i; /* skip warning about unused parameter */
-    /* empty */
-}
-
 #endif /* USE_UCONTEXT || USE_FORK */
 
 #ifdef USE_UCONTEXT
@@ -88,12 +80,11 @@ unsigned long stunnel_thread_id(void) {
 }
 
 NOEXPORT CONTEXT *new_context(void) {
-    static int next_id=1;
+    static unsigned long next_id=1;
     CONTEXT *context;
 
     /* allocate and fill the CONTEXT structure */
-    context=str_alloc(sizeof(CONTEXT));
-    str_detach(context);
+    context=str_alloc_detached(sizeof(CONTEXT));
     context->id=next_id++;
     context->fds=NULL;
     context->ready=0;
@@ -115,12 +106,14 @@ int sthreads_init(void) {
         s_log(LOG_ERR, "Cannot create the listening context");
         return 1;
     }
+    /* update tls for newly allocated ready_head */
+    ui_tls=tls_alloc(NULL, ui_tls, "ui");
     /* no need to initialize ucontext_t structure here
        it will be initialied with swapcontext() call */
     return 0;
 }
 
-int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
+int create_client(SOCKET ls, SOCKET s, CLI *arg, void *(*cli)(void *)) {
     CONTEXT *context;
 
     (void)ls; /* this parameter is only used with USE_FORK */
@@ -128,8 +121,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     s_log(LOG_DEBUG, "Creating a new context");
     context=new_context();
     if(!context) {
-        if(arg)
-            str_free(arg);
+        str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
@@ -138,8 +130,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     /* initialize context_t structure */
     if(getcontext(&context->context)<0) {
         str_free(context);
-        if(arg)
-            str_free(arg);
+        str_free(arg);
         if(s>=0)
             closesocket(s);
         ioerror("getcontext");
@@ -148,8 +139,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     context->context.uc_link=NULL; /* stunnel does not use uc_link */
 
     /* create stack */
-    context->stack=str_alloc(arg->opt->stack_size);
-    str_detach(context->stack);
+    context->stack=str_alloc_detached(arg->opt->stack_size);
 #if defined(__sgi) || ARGC==2 /* obsolete ss_sp semantics */
     context->context.uc_stack.ss_sp=context->stack+arg->opt->stack_size-8;
 #else
@@ -180,15 +170,14 @@ unsigned long stunnel_thread_id(void) {
 }
 
 NOEXPORT void null_handler(int sig) {
-    (void)sig; /* skip warning about unused parameter */
+    (void)sig; /* squash the unused parameter warning */
     signal(SIGCHLD, null_handler);
 }
 
-int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
+int create_client(SOCKET ls, SOCKET s, CLI *arg, void *(*cli)(void *)) {
     switch(fork()) {
     case -1:    /* error */
-        if(arg)
-            str_free(arg);
+        str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
@@ -199,8 +188,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
         cli(arg);
         _exit(0);
     default:    /* parent */
-        if(arg)
-            str_free(arg);
+        str_free(arg);
         if(s>=0)
             closesocket(s);
     }
@@ -211,96 +199,96 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
 
 #ifdef USE_PTHREAD
 
-static pthread_mutex_t stunnel_cs[CRIT_SECTIONS];
-static pthread_mutex_t *lock_cs;
-
-void enter_critical_section(SECTION_CODE i) {
-    pthread_mutex_lock(stunnel_cs+i);
-}
-
-void leave_critical_section(SECTION_CODE i) {
-    pthread_mutex_unlock(stunnel_cs+i);
-}
-
-NOEXPORT void locking_callback(int mode, int type, const char *file, int line) {
-    (void)file; /* skip warning about unused parameter */
-    (void)line; /* skip warning about unused parameter */
-    if(mode&CRYPTO_LOCK)
-        pthread_mutex_lock(lock_cs+type);
-    else
-        pthread_mutex_unlock(lock_cs+type);
-}
-
 struct CRYPTO_dynlock_value {
-    pthread_mutex_t mutex;
+    pthread_rwlock_t rwlock;
 };
+static struct CRYPTO_dynlock_value *lock_cs;
 
 NOEXPORT struct CRYPTO_dynlock_value *dyn_create_function(const char *file,
         int line) {
     struct CRYPTO_dynlock_value *value;
 
-    (void)file; /* skip warning about unused parameter */
-    (void)line; /* skip warning about unused parameter */
-    value=str_alloc(sizeof(struct CRYPTO_dynlock_value));
-    str_detach(value);
-    pthread_mutex_init(&value->mutex, NULL);
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+    value=str_alloc_detached(sizeof(struct CRYPTO_dynlock_value));
+    pthread_rwlock_init(&value->rwlock, NULL);
     return value;
 }
 
 NOEXPORT void dyn_lock_function(int mode, struct CRYPTO_dynlock_value *value,
         const char *file, int line) {
-    (void)file; /* skip warning about unused parameter */
-    (void)line; /* skip warning about unused parameter */
-    if(mode&CRYPTO_LOCK)
-        pthread_mutex_lock(&value->mutex);
-    else
-        pthread_mutex_unlock(&value->mutex);
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+    if(mode&CRYPTO_LOCK) {
+        /* either CRYPTO_READ or CRYPTO_WRITE (but not both) are needed */
+        if(!(mode&CRYPTO_READ)==!(mode&CRYPTO_WRITE))
+            fatal("Invalid locking mode");
+        if(mode&CRYPTO_WRITE)
+            pthread_rwlock_wrlock(&value->rwlock);
+        else
+            pthread_rwlock_rdlock(&value->rwlock);
+    } else
+        pthread_rwlock_unlock(&value->rwlock);
 }
 
 NOEXPORT void dyn_destroy_function(struct CRYPTO_dynlock_value *value,
         const char *file, int line) {
-    (void)file; /* skip warning about unused parameter */
-    (void)line; /* skip warning about unused parameter */
-    pthread_mutex_destroy(&value->mutex);
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+    pthread_rwlock_destroy(&value->rwlock);
     str_free(value);
 }
 
+NOEXPORT void locking_callback(int mode, int type, const char *file, int line) {
+    dyn_lock_function(mode, lock_cs+type, file, line);
+}
+
 unsigned long stunnel_process_id(void) {
     return (unsigned long)getpid();
 }
 
 unsigned long stunnel_thread_id(void) {
 #if defined(SYS_gettid) && defined(__linux__)
-    return syscall(SYS_gettid);
+    return (unsigned long)syscall(SYS_gettid);
 #else
     return (unsigned long)pthread_self();
 #endif
 }
 
+#if OPENSSL_VERSION_NUMBER>=0x10000000L
+NOEXPORT void threadid_func(CRYPTO_THREADID *tid) {
+    CRYPTO_THREADID_set_numeric(tid, stunnel_thread_id());
+}
+#endif
+
 int sthreads_init(void) {
     int i;
 
-    /* initialize stunnel critical sections */
-    for(i=0; i<CRIT_SECTIONS; i++)
-        pthread_mutex_init(stunnel_cs+i, NULL);
-
-    /* initialize OpenSSL locking callback */
-    lock_cs=str_alloc(CRYPTO_num_locks()*sizeof(pthread_mutex_t));
-    str_detach(lock_cs); /* do not track this allocation */
-    for(i=0; i<CRYPTO_num_locks(); i++)
-        pthread_mutex_init(lock_cs+i, NULL);
-    CRYPTO_set_id_callback(stunnel_thread_id);
-    CRYPTO_set_locking_callback(locking_callback);
-
     /* initialize OpenSSL dynamic locks callbacks */
     CRYPTO_set_dynlock_create_callback(dyn_create_function);
     CRYPTO_set_dynlock_lock_callback(dyn_lock_function);
     CRYPTO_set_dynlock_destroy_callback(dyn_destroy_function);
 
+    /* initialize stunnel critical sections */
+    for(i=0; i<STUNNEL_LOCKS; i++)
+        stunnel_locks[i]=CRYPTO_get_new_dynlockid();
+
+    /* initialize OpenSSL locking callback */
+    lock_cs=str_alloc_detached(
+        (size_t)CRYPTO_num_locks()*sizeof(struct CRYPTO_dynlock_value));
+    for(i=0; i<CRYPTO_num_locks(); i++)
+        pthread_rwlock_init(&lock_cs[i].rwlock, NULL);
+#if OPENSSL_VERSION_NUMBER>=0x10000000L
+    CRYPTO_THREADID_set_callback(threadid_func);
+#else
+    CRYPTO_set_id_callback(stunnel_thread_id);
+#endif
+    CRYPTO_set_locking_callback(locking_callback);
+
     return 0;
 }
 
-int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
+int create_client(SOCKET ls, SOCKET s, CLI *arg, void *(*cli)(void *)) {
     pthread_t thread;
     pthread_attr_t pth_attr;
     int error;
@@ -331,8 +319,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     if(error) {
         errno=error;
         ioerror("pthread_create");
-        if(arg)
-            str_free(arg);
+        str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
@@ -344,46 +331,30 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
 
 #ifdef USE_WIN32
 
-static CRITICAL_SECTION stunnel_cs[CRIT_SECTIONS];
-static CRITICAL_SECTION *lock_cs;
-
-void enter_critical_section(SECTION_CODE i) {
-    EnterCriticalSection(stunnel_cs+i);
-}
-
-void leave_critical_section(SECTION_CODE i) {
-    LeaveCriticalSection(stunnel_cs+i);
-}
-
-NOEXPORT void locking_callback(int mode, int type, const char *file, int line) {
-    (void)file; /* skip warning about unused parameter */
-    (void)line; /* skip warning about unused parameter */
-    if(mode&CRYPTO_LOCK)
-        EnterCriticalSection(lock_cs+type);
-    else
-        LeaveCriticalSection(lock_cs+type);
-}
+/* Slim Reader/Writer (SRW) Lock would be better than CRITICAL_SECTION,
+ * but it is unsupported on Windows XP (and earlier versions of Windows):
+ * https://msdn.microsoft.com/en-us/library/windows/desktop/aa904937%28v=vs.85%29.aspx */
 
 struct CRYPTO_dynlock_value {
     CRITICAL_SECTION mutex;
 };
+static struct CRYPTO_dynlock_value *lock_cs;
 
 NOEXPORT struct CRYPTO_dynlock_value *dyn_create_function(const char *file,
         int line) {
     struct CRYPTO_dynlock_value *value;
 
-    (void)file; /* skip warning about unused parameter */
-    (void)line; /* skip warning about unused parameter */
-    value=str_alloc(sizeof(struct CRYPTO_dynlock_value));
-    str_detach(value);
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+    value=str_alloc_detached(sizeof(struct CRYPTO_dynlock_value));
     InitializeCriticalSection(&value->mutex);
     return value;
 }
 
 NOEXPORT void dyn_lock_function(int mode, struct CRYPTO_dynlock_value *value,
         const char *file, int line) {
-    (void)file; /* skip warning about unused parameter */
-    (void)line; /* skip warning about unused parameter */
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
     if(mode&CRYPTO_LOCK)
         EnterCriticalSection(&value->mutex);
     else
@@ -392,12 +363,16 @@ NOEXPORT void dyn_lock_function(int mode, struct CRYPTO_dynlock_value *value,
 
 NOEXPORT void dyn_destroy_function(struct CRYPTO_dynlock_value *value,
         const char *file, int line) {
-    (void)file; /* skip warning about unused parameter */
-    (void)line; /* skip warning about unused parameter */
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
     DeleteCriticalSection(&value->mutex);
     str_free(value);
 }
 
+NOEXPORT void locking_callback(int mode, int type, const char *file, int line) {
+    dyn_lock_function(mode, lock_cs+type, file, line);
+}
+
 unsigned long stunnel_process_id(void) {
     return GetCurrentProcessId() & 0x00ffffff;
 }
@@ -409,33 +384,33 @@ unsigned long stunnel_thread_id(void) {
 int sthreads_init(void) {
     int i;
 
-    /* initialize stunnel critical sections */
-    for(i=0; i<CRIT_SECTIONS; i++)
-        InitializeCriticalSection(stunnel_cs+i);
-
-    /* initialize OpenSSL locking callback */
-    lock_cs=str_alloc(CRYPTO_num_locks()*sizeof(CRITICAL_SECTION));
-    str_detach(lock_cs); /* do not track this allocation */
-    for(i=0; i<CRYPTO_num_locks(); i++)
-        InitializeCriticalSection(lock_cs+i);
-    CRYPTO_set_locking_callback(locking_callback);
-
     /* initialize OpenSSL dynamic locks callbacks */
     CRYPTO_set_dynlock_create_callback(dyn_create_function);
     CRYPTO_set_dynlock_lock_callback(dyn_lock_function);
     CRYPTO_set_dynlock_destroy_callback(dyn_destroy_function);
 
+    /* initialize stunnel critical sections */
+    for(i=0; i<STUNNEL_LOCKS; i++)
+        stunnel_locks[i]=CRYPTO_get_new_dynlockid();
+
+    /* initialize OpenSSL locking callback */
+    lock_cs=str_alloc_detached(
+        (size_t)CRYPTO_num_locks()*sizeof(struct CRYPTO_dynlock_value));
+    for(i=0; i<CRYPTO_num_locks(); i++)
+        InitializeCriticalSection(&lock_cs[i].mutex);
+    CRYPTO_set_locking_callback(locking_callback);
+
     return 0;
 }
 
-int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
+int create_client(SOCKET ls, SOCKET s, CLI *arg, void *(*cli)(void *)) {
     (void)ls; /* this parameter is only used with USE_FORK */
     s_log(LOG_DEBUG, "Creating a new thread");
-    if((long)_beginthread((void(*)(void *))cli, arg->opt->stack_size, arg)==-1) {
+    if((long)_beginthread((void(*)(void *))cli,
+            (unsigned)arg->opt->stack_size, arg)==-1) {
         ioerror("_beginthread");
-        if(arg)
-            str_free(arg);
-        if(s>=0)
+        str_free(arg);
+        if(s!=INVALID_SOCKET)
             closesocket(s);
         return -1;
     }
@@ -447,14 +422,6 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
 
 #ifdef USE_OS2
 
-void enter_critical_section(SECTION_CODE i) {
-    DosEnterCritSec();
-}
-
-void leave_critical_section(SECTION_CODE i) {
-    DosExitCritSec();
-}
-
 int sthreads_init(void) {
     return 0;
 }
@@ -471,13 +438,12 @@ unsigned long stunnel_thread_id(void) {
     return (unsigned long)ppib->pib_ulpid;
 }
 
-int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
+int create_client(SOCKET ls, SOCKET s, CLI *arg, void *(*cli)(void *)) {
     (void)ls; /* this parameter is only used with USE_FORK */
     s_log(LOG_DEBUG, "Creating a new thread");
     if((long)_beginthread((void(*)(void *))cli, NULL, arg->opt->stack_size, arg)==-1L) {
         ioerror("_beginthread");
-        if(arg)
-            str_free(arg);
+        str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
@@ -513,12 +479,12 @@ void _endthread(void) {
 #ifdef DEBUG_STACK_SIZE
 
 #define STACK_RESERVE (STACK_SIZE/8)
-#define VERIFY_AREA ((STACK_SIZE-STACK_RESERVE)/sizeof(u32))
+#define VERIFY_AREA ((STACK_SIZE-STACK_RESERVE)/sizeof(uint32_t))
 #define TEST_VALUE 0xdeadbeef
 
 /* some heuristic to determine the usage of client stack size */
 void stack_info(int init) { /* 1-initialize, 0-display */
-    u32 table[VERIFY_AREA];
+    uint32_t table[VERIFY_AREA];
     int i, num;
     static int min_num=VERIFY_AREA;
 
@@ -546,10 +512,10 @@ void stack_info(int init) { /* 1-initialize, 0-display */
         s_log(LOG_NOTICE,
             "stack_info: size=%d, current=%d (%d%%), maximum=%d (%d%%)",
             STACK_SIZE,
-            (int)((VERIFY_AREA-num)*sizeof(u32)),
-            (int)((VERIFY_AREA-num)*sizeof(u32)*100/STACK_SIZE),
-            (int)((VERIFY_AREA-min_num)*sizeof(u32)),
-            (int)((VERIFY_AREA-min_num)*sizeof(u32)*100/STACK_SIZE));
+            (int)((VERIFY_AREA-num)*sizeof(uint32_t)),
+            (int)((VERIFY_AREA-num)*sizeof(uint32_t)*100/STACK_SIZE),
+            (int)((VERIFY_AREA-min_num)*sizeof(uint32_t)),
+            (int)((VERIFY_AREA-min_num)*sizeof(uint32_t)*100/STACK_SIZE));
     }
 }
 
